from cmath import sqrt
import numpy as np
import pandas as pd
import polars as pl
from scipy import stats
from typing import Union
import seaborn as sns
from statsmodels.stats.proportion import proportions_ztest
from statsmodels.stats import multitest
import pyrsm.basics.utils as bu
import pyrsm.radiant.utils as ru
from pyrsm.model.model import sig_stars
from pyrsm.utils import ifelse, check_dataframe


class compare_props:
    """
    Compare proportions across levels of a categorical variable in a Pandas
    or Polars dataframe. See the notebook linked below for a worked example,
    including the web UI:

    https://github.com/vnijs/pyrsm/blob/main/examples/basics-compare-means.ipynb
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        var1: str,
        var2: str,
        lev: str,
        comb: list[tuple[str, str]] = [],
        alt_hyp: str = "two-sided",
        conf: float = 0.95,
        adjust: str = None,
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"

        self.data = check_dataframe(self.data)
        self.var1 = var1
        self.var2 = var2
        self.lev = lev
        # self.comb = comb
        self.alt_hyp = alt_hyp
        self.conf = conf
        self.alpha = 1 - self.conf
        self.adjust = adjust

        self.data[self.var1] = self.data[self.var1].astype("category")
        self.levels = list(self.data[self.var1].cat.categories)

        if len(comb) == 0:
            self.comb = ru.iterms(self.levels)
        else:
            self.comb = ifelse(isinstance(comb, str), [comb], comb)

        descriptive_stats = []
        for lev in self.levels:
            subset = self.data.loc[
                (self.data[self.var2] == self.lev) & (self.data[self.var1] == lev),
                self.var2,
            ]
            ns = len(subset)
            n_missing = subset.isna().sum()
            n = len(self.data.loc[(self.data[self.var1] == lev), self.var2]) - n_missing
            p = ns / n
            sd = sqrt(p * (1 - p)).real
            se = (sd / sqrt(n)).real
            z_score = stats.norm.ppf(1 - self.alpha / 2)
            me = (z_score * sd / sqrt(n)).real
            descriptive_stats.append([lev, ns, p, n, n_missing, sd, se, me])

        self.descriptive_stats = pd.DataFrame(
            descriptive_stats,
            columns=[
                self.var1,
                self.lev,
                "p",
                "n",
                "n_missing",
                "sd",
                "se",
                "me",
            ],
        )

        if self.alt_hyp == "less":
            alt_hyp_sign = "less than"
        elif self.alt_hyp == "two-sided":
            alt_hyp_sign = "not equal to"
        else:
            alt_hyp_sign = "greater than"

        def wald_ci(n1, p1, n2, p2, z):
            # what R uses for a comparison of proportions
            se = np.sqrt((p1 * (1 - p1) / n1) + (p2 * (1 - p2) / n2))
            diff = p1 - p2
            return [diff - z * se, diff + z * se]

        comp_stats = []
        for c in self.comb:
            v1, v2 = c.split(":")
            null_hyp = f"{v1} = {v2}"
            alt_hyp = f"{v1} {alt_hyp_sign} {v2}"

            x = self.data.loc[self.data[self.var1] == v1, self.var2].dropna()
            y = self.data.loc[self.data[self.var1] == v2, self.var2].dropna()

            c1 = np.sum(x == self.lev)
            c2 = np.sum(y == self.lev)
            n1 = len(x)
            n2 = len(y)
            p1 = c1 / n1
            p2 = c2 / n2
            diff = p1 - p2

            pzt = ifelse(
                self.alt_hyp == "less",
                "smaller",
                ifelse(self.alt_hyp == "greater", "larger", "two-sided"),
            )

            z_val, p_val = proportions_ztest([c1, c2], [n1, n2], alternative=pzt)
            zc = stats.norm.ppf(self.conf)
            if self.alt_hyp == "less":
                ci = [-1, wald_ci(n1, p1, n2, p2, zc)[1]]
            elif self.alt_hyp == "two-sided":
                zc = stats.norm.ppf(1 - (1 - self.conf) / 2)
                ci = wald_ci(n1, p1, n2, p2, zc)
            else:
                ci = [wald_ci(n1, p1, n2, p2, zc)[0], 1]

            comp_stats.append(
                [
                    null_hyp,
                    alt_hyp,
                    diff,
                    p_val,
                    z_val**2,
                    1,
                    ci[0],
                    ci[1],
                    sig_stars([p_val])[0],
                ]
            )

        cl = bu.ci_label(self.alt_hyp, self.conf)
        self.comp_stats = pd.DataFrame(
            comp_stats,
            columns=[
                "Null hyp.",
                "Alt. hyp.",
                "diff",
                "p.value",
                "chisq.value",
                "df",
                cl[0],
                cl[1],
                "",
            ],
        )

        if self.adjust is not None:
            if self.alt_hyp == "two-sided":
                alpha = self.alpha
            else:
                alpha = self.alpha * 2
            self.comp_stats["p.value"] = multitest.multipletests(
                self.comp_stats["p.value"], method=self.adjust, alpha=alpha
            )[1]
            self.comp_stats[""] = sig_stars(self.comp_stats["p.value"])

    def summary(self, extra=False, dec: int = 3) -> None:
        comp_stats = self.comp_stats.copy()
        if not extra:
            comp_stats = comp_stats.iloc[:, [0, 1, 2, 3, -1]]

        comp_stats["p.value"] = ifelse(
            comp_stats["p.value"] < 0.001, "< .001", round(comp_stats["p.value"], dec)
        )
        print("Pairwise proportion comparisons")
        print(f"Data      : {self.name}")
        print(f"Variables : {self.var1}, {self.var2}")
        print(f'Level     : "{self.lev}" in {self.var2}')
        print(f"Confidence: {self.conf}")
        print(f"Adjustment: {self.adjust}\n")
        print(self.descriptive_stats.round(dec).to_string(index=False), "\n")
        print(comp_stats.round(dec).to_string(index=False))
        print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")

    def plot(self, plots: str = "bar") -> None:
        if plots == "bar":
            data = self.data[[self.var1, self.var2]].copy()
            data[self.var2] = ifelse(data[self.var2] == self.lev, 1.0, 0.0)
            sns.barplot(
                data=data,
                y=self.var2,
                x=self.var1,
                # yerr=self.descriptive_stats["se"], weird issue with se's ...
            )
        elif plots == "dodge":
            data = self.data[[self.var1, self.var2]].copy()
            pt = (
                data.groupby(self.var1, observed=False)
                .value_counts(normalize=True)
                .reset_index()
            )
            sns.barplot(
                data=pt,
                y="proportion",
                x=self.var1,
                hue=self.var2,
                dodge=True,
            )
        else:
            print("Invalid plot type")



import pandas as pd
import numpy as np
from shiny import App, Inputs, Outputs, Session, reactive, render, ui


# Create a date range
dates = pd.date_range("20230101", periods=6)

# Create a DataFrame
data = pd.DataFrame(
    {
        "A": np.random.rand(6),  # numeric column
        "B": pd.Categorical(
            ["test", "train", "test", "train", "test", "train"]
        ),  # categorical column
        "C": dates,  # date column
    }
)

app_ui = ui.page_fluid(
    ui.input_select(
        "selection_mode",
        "Selection mode",
        {"none": "(None)", "single": "Single", "multiple": "Multiple"},
        selected="multiple",
    ),
    ui.input_switch("gridstyle", "Grid", True),
    ui.input_switch("fullwidth", "Take full width", True),
    ui.input_switch("fixedheight", "Fixed height", True),
    ui.output_data_frame("grid"),
    ui.panel_fixed(
        ui.output_text_verbatim("detail"),
        right="10px",
        bottom="10px",
    ),
    class_="p-3",
)


def server(input: Inputs, output: Outputs, session: Session):
    df: reactive.Value[pd.DataFrame] = reactive.Value(data)

    @output
    @render.data_frame
    def grid():
        height = 350 if input.fixedheight() else None
        width = "100%" if input.fullwidth() else "fit-content"
        if input.gridstyle():
            return render.DataGrid(
                df(),
                row_selection_mode=input.selection_mode(),
                height=height,
                width=width,
            )
        else:
            return render.DataTable(
                df(),
                row_selection_mode=input.selection_mode(),
                height=height,
                width=width,
            )

    @reactive.Effect
    @reactive.event(input.grid_cell_edit)
    def handle_edit():
        edit = input.grid_cell_edit()
        df_copy = df().copy()
        df_copy.iat[edit["row"], edit["col"]] = edit["new_value"]
        df.set(df_copy)

    @output
    @render.text
    def detail():
        if (
            input.grid_selected_rows() is not None
            and len(input.grid_selected_rows()) > 0
        ):
            # "split", "records", "index", "columns", "values", "table"
            return df().iloc[list(input.grid_selected_rows())]


app = App(app_ui, server)



import streamlit as st
import pandas as pd
import pyrsm as rsm
import io
from contextlib import redirect_stdout
from datetime import datetime

st.set_page_config(layout="wide")
st.set_option("deprecation.showPyplotGlobalUse", False)

tab1, tab2, tab3 = st.tabs(["Data", "Summary", "Plot"])


def view_df(df):
    st.dataframe(df)


def describe_df(df):
    if hasattr(df, "description"):
        st.markdown(df.description, unsafe_allow_html=True)


@st.cache_resource
def logistic_regression(df, X, y):
    return rsm.model.logistic(dataset=df, rvar=y, evar=X)


# @st.cache_data
def logistic_summary(_lr):
    out = io.StringIO()
    with redirect_stdout(out):
        _lr.summary()
    dt_string = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    st.write(f"Date and time: {dt_string}")
    return out.getvalue()


# @st.cache_data
def logistic_plot(_lr, plots="or"):
    return _lr.plot(plots=plots)


@st.cache_data
def get_df(file):
    fname = file.name.split(".")[0]
    extension = file.name.split(".")[-1].upper()
    if extension == "CSV":
        df = pd.read_csv(file)
        code = f"""{fname} = pd.read_csv({file.name})"""
    elif extension in ["XLS", "XLSX"]:
        df = pd.read_excel(file, engine="openpyxl")
        code = f"""{fname} = pd.read_excel("{file.name}", engine="openpyxl")"""
    elif extension in ["PKL", "PICKLE"]:
        df = pd.read_pickle(file)
        code = f"""{fname} = pd.read_pickle("{file.name}")"""
    return df, fname, code


def main():

    file = st.sidebar.file_uploader("Upload file", type=["csv", "xlsx", "pkl"])
    if not file:
        st.sidebar.write("Upload a .csv, .xlsx or .pkl file to get started")
        # need return and a 'main' function to avoid actions the
        # rest of the code is not ready for yet
        return None

    df, fname, code = get_df(file)

    with tab1:
        view_df(df)
        describe_df(df)

    df_cols = df.columns.tolist()
    rvar = st.sidebar.selectbox("Response variable:", df.columns)
    df_cols.remove(rvar)
    evar = st.sidebar.multiselect("Explanatory variables:", df_cols)

    if len(evar) > 0:
        lr = logistic_regression(df, evar, rvar)
    else:
        lr = None

    with tab2:
        if lr:
            # st.text(logistic_summary(lr))
            st.code(logistic_summary(lr))
            st.code(
                f"""{code}\nlr = rsm.model.logistic(dataset={fname}, rvar="{rvar}", evar={evar})\nlr.summary()"""
            )

    with tab3:
        plot_types = {
            "Odds ratio": "or",
            "Prediction": "pred",
            "Variable important": "vimp",
        }
        plots = st.sidebar.selectbox("Plot types:", plot_types)
        if lr:
            st.pyplot(logistic_plot(lr, plots=plot_types[plots]))
            st.code(
                f"""{code}\nlr = rsm.logitic(dataset={fname}, rvar="{rvar}", evar={evar})\nlr.plot(plots="{plot_types[plots]}")"""
            )


main()

# to run this app use the below from a terminal
# then open a browser at the url localhost:8501
# streamlit run streamlit-logistic-app.py



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pyrsm'
copyright = "2025, Vincent Nijs"
author = 'Vincent Nijs'
release = "2025"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']



from .data.data_view import *
from .basics.probability_calculator import *
from .basics.single_mean import *
from .basics.compare_means import *
from .basics.single_prop import *
from .basics.compare_props import *
from .basics.cross_tabs import *
from .basics.goodness import *
from .basics.correlation import *
from .model.regress import *
from .model.logistic import *
from .model.rforest import *
from .model.mlp import *
from .model.xgboost import *
from .radiant import *



import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import seaborn as sns

# from pyrsm import xtile, bincode
from ..bins import xtile, bincode
from ..utils import ifelse, table2data
from sklearn import metrics
from scipy.stats import rankdata


def calc_qnt(df, rvar, lev, pred, qnt=10):
    """
    Create quantiles and calculate input to use for lift and gains charts

    Parameters
    ----------
    df : Pandas dataframe
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    qnt : int
        Number of quantiles to create

    Returns
    -------
    Pandas dataframe
        Response metrics per quantile. Used as input for lift and gains charts
    """

    df = df.loc[:, (rvar, pred)]
    df["bins"] = xtile(df[pred], qnt)
    df["rvar_int"] = ifelse(df[rvar] == lev, 1, ifelse(df[rvar].isna(), np.nan, 0))
    perf_df = (
        df.groupby("bins", observed=False)
        .rvar_int.agg(nr_obs="count", nr_resp="sum")
        .reset_index()
    )

    # flip if needed
    if perf_df.nr_resp.iloc[1] < perf_df.nr_resp.iloc[-1]:
        perf_df = perf_df.sort_values("bins", ascending=False)

    perf_df["cum_obs"] = np.cumsum(perf_df.nr_obs)
    perf_df["cum_prop"] = perf_df.cum_obs / perf_df.cum_obs.iloc[-1]
    perf_df["cum_resp"] = np.cumsum(perf_df.nr_resp)
    return perf_df


def gains_tab(df, rvar, lev, pred, qnt=10):
    """
    Calculate cumulative gains using the cum_resp column

    Parameters
    ----------
    df : Pandas dataframe
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    qnt : int
        Number of quantiles to create

    Returns
    -------
    Pandas dataframe
        Gains measures per quantile. Input for gains chart
    """

    df = calc_qnt(df, rvar, lev, pred, qnt=qnt)
    df["cum_gains"] = df.cum_resp / df.cum_resp.iloc[-1]
    df0 = pd.DataFrame({"cum_prop": [0], "cum_gains": [0]})
    df = pd.concat([df0, df], sort=False).reset_index(drop=True)
    df.index = range(df.shape[0])
    return df[["cum_prop", "cum_gains"]]


def lift_tab(df, rvar, lev, pred, qnt=10):
    """
    Calculate cumulative lift using the cum_resp and the cum_obs column

    Parameters
    ----------
    df : Pandas dataframe
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    qnt : int
        Number of quantiles to create

    Returns
    -------
    Pandas dataframe
        Lift measures per quantile. Input for lift chart
    """

    df = calc_qnt(df, rvar, lev, pred, qnt=qnt)
    df["cum_resp_rate"] = df.cum_resp / df.cum_obs
    df["cum_lift"] = df.cum_resp_rate / df.cum_resp_rate.iloc[-1]
    df.index = range(df.shape[0])
    return df[["cum_prop", "cum_lift"]]


def confusion(df, rvar, lev, pred, cost=1, margin=2):
    """
    Calculate TP, FP, TN, FN, and contact

    Parameters
    ----------
    df : Pandas dataframe
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action

    Returns
    -------
    TP : int
        Number of True Positive predictions
    FP : int
        Number of False Positive predictions
    TN : int
        Number of True Negative predictions
    FN : int
        Number of False Negative predictions
    contact: float
        Proportion of cases to act on based on the cost/margin ratio
    """

    if pd.api.types.is_list_like(pred) and len(pred) > 1:
        return "This function can only take one predictor variables at time"

    break_even = cost / margin
    gtbe = df[pred] > break_even
    pos = df[rvar] == lev
    TP = np.where(gtbe & pos, 1, 0).sum()
    FP = np.where(gtbe & np.logical_not(pos), 1, 0).sum()
    TN = np.where(np.logical_not(gtbe) & np.logical_not(pos), 1, 0).sum()
    FN = np.where(np.logical_not(gtbe) & pos, 1, 0).sum()
    contact = (TP + FP) / (TP + FP + TN + FN)
    return TP, FP, TN, FN, contact


def uplift_tab(df, rvar, lev, pred, tvar, tlev, scale=1, qnt=10):
    """
    Calculate an Uplift table

    Parameters
    ----------
    df : Pandas dataframe
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    tvar : str
        Name of the treatment variable column in df
    tlev : str
        Name of the 'success' level in tvar
    scale : float
        Scaling factor to use in calculations
    qnt : int
        Number of quantiles to create

    Returns
    -------
    Pandas dataframe
        Incremental uplift per quantile. Input for uplift charts
    """

    def local_xtile(x, treatment, n=qnt, rev=True):
        x = np.array(x)
        treatment = np.array(treatment)
        breaks = np.concatenate(
            (
                np.array([-np.inf]),
                np.quantile(x[treatment], np.arange(0, n + 1) / n, method="linear")[
                    1:-1
                ],
                np.array([np.inf]),
            )
        )

        if len(np.unique(breaks)) == len(breaks):
            bins = pd.cut(x, breaks, include_lowest=True, labels=False) + 1
        else:
            bins = bincode(x, breaks)

        if rev is True:
            bins = (n + 1) - bins

        return bins

    df = df.copy()
    df[rvar] = df[rvar] == lev
    df[tvar] = df[tvar] == tlev

    df["bins"] = local_xtile(df[pred], df[tvar], n=qnt, rev=True)
    df["T_resp"] = df[tvar] & df[rvar]
    df["C_resp"] = ~df[tvar] & df[rvar]
    df["C_n"] = ~df[tvar]
    tab = (
        (
            (
                df.groupby("bins", observed=False).agg(
                    nr_obs=("bins", "count"),
                    nr_resp=(rvar, "sum"),
                    T_resp=("T_resp", "sum"),
                    T_n=(tvar, "sum"),
                    C_resp=("C_resp", "sum"),
                    C_n=("C_n", "sum"),
                )
            )
            .assign(
                uplift=lambda x: x.T_resp / x.T_n - x.C_resp / x.C_n,
                cum_prop=lambda x: x.index / qnt,
                T_resp=lambda x: x.T_resp.cumsum() * scale,
                T_n=lambda x: x.T_n.cumsum() * scale,
                C_resp=lambda x: x.C_resp.cumsum() * scale,
                C_n=lambda x: x.C_n.cumsum() * scale,
                incremental_resp=lambda x: x.T_resp - x.C_resp * x.T_n / x.C_n,
            )
            .assign(
                inc_uplift=lambda x: x.incremental_resp / x.T_n.max() * 100, pred=pred
            )
        )
        .reset_index()
        .loc[
            :,
            [
                "pred",
                "bins",
                "cum_prop",
                "T_resp",
                "T_n",
                "C_resp",
                "C_n",
                "incremental_resp",
                "inc_uplift",
                "uplift",
            ],
        ]
    )
    return tab


def inc_uplift_plot(
    df, rvar, lev, pred, tvar, tlev, scale=1, qnt=10, marker="o", **kwargs
):
    """
    Plot an Incremental Uplift chart

    Parameters
    ----------
    df : Pandas dataframe or a dictionary of dataframes with keys to show multiple curves for different models or data samples
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    tvar : str
        Name of the treatment variable column in df
    tlev : str
        Name of the 'success' level in tvar
    scale : float
        Scaling factor to use in calculations
    qnt : int
        Number of quantiles to create
    **kwargs : Named arguments to be passed to the seaborn lineplot function

    Returns
    -------
    Seaborn object
        Plot of Incremental Uplift per quantile
    """
    dct = ifelse(isinstance(df, dict), df, {"": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)
    group = ifelse(len(pred) > 1 or len(dct.keys()) > 1, "pred", None)

    rd = [
        pd.concat(
            [
                pd.DataFrame({"cum_prop": [0], "inc_uplift": [0], "pred": p}),
                uplift_tab(
                    dct[k], rvar, lev, p, tvar, tlev, scale=scale, qnt=qnt
                ).assign(predictor=p + ifelse(k == "", k, f" ({k})"))[
                    ["cum_prop", "inc_uplift", "pred"]
                ],
            ]
        )
        for k in dct.keys()
        for p in pred
    ]

    yend = rd[0].inc_uplift.iloc[-1]
    rd = pd.concat(rd).reset_index(drop=True)
    fig = sns.lineplot(
        x="cum_prop", y="inc_uplift", data=rd, hue=group, marker=marker, **kwargs
    )
    fig.yaxis.set_major_formatter(
        mtick.FuncFormatter(lambda y, _: "{:.0%}".format(y / 100))
    )
    fig.xaxis.set_major_formatter(mtick.FuncFormatter(lambda x, _: "{:.0%}".format(x)))
    fig.set(ylabel="Incremental Uplift", xlabel="Percentage of population targeted")
    plt.plot([0, 1], [0, yend], linestyle="--", linewidth=1, color=plt.cm.Blues(0.7))
    if len(dct) > 1 or len(pred) > 1:
        fig.legend(title=None)
    return fig


def inc_profit_tab(
    df,
    rvar,
    lev,
    pred,
    tvar,
    tlev,
    cost=1,
    margin=2,
    scale=1,
    qnt=10,
):
    """
    Tabulate Incremental Profit for Uplift modeling

    Parameters
    ----------
    df : Pandas dataframe or a dictionary of dataframes with keys to show multiple curves for different models or data samples
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    tvar : str
        Name of the treatment variable column in df
    tlev : str
        Name of the 'success' level in tvar
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    scale : float
        Scaling factor to use in calculations
    qnt : int
        Number of quantiles to create

    Returns
    -------
    Seaborn object
        Plot of Incremental Uplift per quantile
    """

    dct = ifelse(isinstance(df, dict), df, {"": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)

    rd = [
        pd.concat(
            [
                pd.DataFrame(
                    {"cum_prop": [0], "incremental_resp": [0], "T_n": [0], "pred": p}
                ),
                uplift_tab(
                    dct[k], rvar, lev, p, tvar, tlev, scale=scale, qnt=qnt
                ).assign(predictor=p + ifelse(k == "", k, f" ({k})"))[
                    ["cum_prop", "incremental_resp", "T_n", "pred"]
                ],
            ]
        )
        for k in dct.keys()
        for p in pred
    ]

    rd = pd.concat(rd).reset_index(drop=True)
    return rd.assign(inc_profit=(rd.incremental_resp * margin - rd.T_n * cost))


def inc_profit_plot(
    df,
    rvar,
    lev,
    pred,
    tvar,
    tlev,
    cost=1,
    margin=2,
    scale=1,
    qnt=10,
    contact=True,
    marker="o",
    prn=False,
    **kwargs,
):
    """
    Plot an Incremental Profit chart for Uplift modeling

    Parameters
    ----------
    df : Pandas dataframe or a dictionary of dataframes with keys to show multiple curves for different models or data samples
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    tvar : str
        Name of the treatment variable column in df
    tlev : str
        Name of the 'success' level in tvar
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    scale : float
        Scaling factor to use in calculations
    qnt : int
        Number of quantiles to create
    contact : bool
        Plot a vertical line that shows the optimal contact level.
    marker : str
        Marker to use for line plot
    prn : bool
        Print the maximum incremental profit value(s) (default is True, requires contact=True)
    **kwargs : Named arguments to be passed to the seaborn lineplot function

    Returns
    -------
    Seaborn object
        Plot of Incremental Profit per quantile
    """

    dct = ifelse(isinstance(df, dict), df, {"": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)
    group = ifelse(len(pred) > 1 or len(dct.keys()) > 1, "predictor", None)

    # rd = inc_profit_tab(dct, rvar, lev, pred, tvar, tlev, cost, margin, scale, qnt)
    # fig = sns.lineplot(
    #     x="cum_prop", y="inc_profit", data=rd, hue=group, marker=marker, **kwargs
    # )

    df = [
        inc_profit_tab(
            dct[k], rvar, lev, p, tvar, tlev, cost, margin, scale, qnt
        ).assign(predictor=p + ifelse(k == "", k, f" ({k})"))
        for k in dct.keys()
        for p in pred
    ]
    df = pd.concat(df).drop(columns="pred").reset_index(drop=True)
    fig = sns.lineplot(
        x="cum_prop", y="inc_profit", data=df, hue=group, marker=marker, **kwargs
    )

    fig.set(ylabel="Incremental Profit", xlabel="Percentage of population targeted")
    fig.yaxis.set_major_formatter(mtick.FuncFormatter(lambda y, _: format(int(y), ",")))
    fig.xaxis.set_major_formatter(mtick.FuncFormatter(lambda x, _: "{:.0%}".format(x)))
    fig.axhline(1, linestyle="--", linewidth=1)
    if contact:
        cnf = [
            confusion(
                dct[k].query(f"{tvar} == {tlev}"),
                rvar,
                lev,
                p,
                cost=cost,
                margin=margin,
            )[-1]
            for k in dct.keys()
            for p in pred
        ]
        if prn:
            print(df)
        [
            [fig.axvline(l, linestyle="--", linewidth=1, color=sns.color_palette()[i])]
            for i, l in enumerate(filter(lambda x: x < 1, cnf))
        ]
    if len(dct) > 1 or len(pred) > 1:
        fig.legend(title=None)

    return fig


def uplift_plot(df, rvar, lev, pred, tvar, tlev, qnt=10, marker="o", **kwargs):
    """
    Plot an Uplift bar chart

    Parameters
    ----------
    df : Pandas dataframe or a dictionary of dataframes with keys to show multiple curves for different models or data samples
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    tvar : str
        Name of the treatment variable column in df
    tlev : str
        Name of the 'success' level in tvar
    qnt : int
        Number of quantiles to create
    **kwargs : Named arguments to be passed to the seaborn lineplot function

    Returns
    -------
    Seaborn object
        Plot of Incremental Uplift per quantile
    """
    dct = ifelse(isinstance(df, dict), df, {"": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)
    group = ifelse(len(pred) > 1 or len(dct.keys()) > 1, "predictor", plt.cm.Blues(0.7))

    rd = [
        uplift_tab(dct[k], rvar, lev, p, tvar, tlev, qnt=qnt).assign(
            predictor=p + ifelse(k == "", k, f" ({k})")
        )
        for k in dct.keys()
        for p in pred
    ]

    rd = pd.concat(rd).reset_index(drop=True)
    if group == "predictor":
        fig = sns.barplot(x="cum_prop", y="uplift", data=rd, hue=group, **kwargs)
    else:
        fig = sns.barplot(x="cum_prop", y="uplift", data=rd, color=group, **kwargs)
    fig.set(ylabel="Uplift", xlabel="Percentage of population targeted")
    fig.yaxis.set_major_formatter(mtick.FuncFormatter(lambda y, _: "{:.0%}".format(y)))
    # fig.xaxis.set_major_formatter(mtick.FuncFormatter(lambda x, _: "{:.0%}".format(x)))

    if len(dct) > 1 or len(pred) > 1:
        fig.legend(title=None)
    return fig


def profit_max(df, rvar, lev, pred, cost=1, margin=2, scale=1):
    """
    Calculate the maximum profit using a dataframe as input

    Parameters
    ----------
    df : Pandas dataframe
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    scale : float
        Scaling factor to use in calculations

    Returns
    -------
    float
        Measure of optimal performance (e.g., profit) based on the specified cost and margin information
    """

    TP, FP, TN, FN, contact = confusion(df, rvar, lev, pred, cost=cost, margin=margin)
    return scale * (margin * TP - cost * (TP + FP))


def profit(rvar, pred, lev=1, cost=1, margin=2, scale=1):
    """
    Calculate the maximum profit using series as input. Provides the same results as profit_max

    Parameters
    ----------
    rvar : Pandas series
        Column from a Pandas dataframe with the response variable
    pred : Pandas series
        Column from a Pandas dataframe with model predictions
    lev : str
        Name of the 'success' level in rvar
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    scale : float
        Scaling factor to use in calculations

    Returns
    -------
    float
        Measure of optimal performace (e.g., profit) based on the specified cost and margin information
    """

    break_even = cost / margin
    TP = ((rvar == lev) & (pred > break_even)).sum()
    FP = ((rvar != lev) & (pred > break_even)).sum()
    return scale * (margin * TP - cost * (TP + FP))


def ROME_max(df, rvar, lev, pred, cost=1, margin=2):
    """
    Calculate the maximum Return on Marketing Expenditures

    Parameters
    ----------
    df : Pandas dataframe
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action

    Returns
    -------
    float
        Measure of optimal performace based on the specified cost and margin information
    """

    TP, FP, TN, FN, contact = confusion(df, rvar, lev, pred, cost=cost, margin=margin)
    profit = margin * TP - cost * (TP + FP)
    return profit / (cost * (TP + FP))


def ROME(pred, rvar, lev, cost=1, margin=2):
    """
    Calculate the maximum Return on Marketing Expenditures using series as input.
    Provides the same results as ROME_max

    Parameters
    ----------
    pred : Pandas series
        Column from a Pandas dataframe with model predictions
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model prediction
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action

    Returns
    -------
    float
        Measure of optimal performace (e.g., profit) based on the specified cost and margin information
    """

    break_even = cost / margin
    TP = ((rvar == lev) & (pred > break_even)).sum()
    FP = ((rvar != lev) & (pred > break_even)).sum()
    profit = margin * TP - cost * (TP + FP)
    return profit / (cost * (TP + FP))


def profit_tab(df, rvar, lev, pred, qnt=10, cost=1, margin=2, scale=1):
    """
    Calculate table with profit per quantile

    Parameters
    ----------
    df : Pandas dataframe
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    qnt : int
        Number of quantiles to create
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    scale : float
        Scaling factor to use in calculations

    Returns
    -------
    Pandas dataframe
        Profit per quantile. Input for profit chart
    """

    df = calc_qnt(df, rvar, lev, pred, qnt=qnt)
    df["cum_profit"] = (margin * df.cum_resp - cost * df.cum_obs) * scale
    df0 = pd.DataFrame({"cum_prop": [0], "cum_profit": [0]})
    df = pd.concat([df0, df], sort=False)
    df.index = range(df.shape[0])
    return df[["cum_prop", "cum_profit"]]


def ROME_tab(df, rvar, lev, pred, qnt=10, cost=1, margin=2):
    """
    Calculate table with Return on Marketing Expenditures per quantile

    Parameters
    ----------
    df : Pandas dataframe
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    qnt : int
        Number of quantiles to create
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action

    Returns
    -------
    Pandas dataframe
        ROME quantile. Input for ROME chart
    """

    df = calc_qnt(df, rvar, lev, pred, qnt=qnt)
    df["cum_profit"] = margin * df.cum_resp - cost * df.cum_obs
    cum_cost = cost * df.cum_obs
    df["ROME"] = (margin * df.cum_resp - cum_cost) / cum_cost
    df.index = range(df.shape[0])
    return df[["cum_prop", "ROME"]]


def profit_plot(
    df,
    rvar,
    lev,
    pred,
    qnt=10,
    cost=1,
    margin=2,
    scale=1,
    contact=True,
    marker="o",
    prn=True,
    **kwargs,
):
    """
    Plot a profit curve

    Parameters
    ----------
    df : Pandas dataframe or a dictionary of dataframes with keys to show multiple curves for different models or data samples
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    qnt : int
        Number of quantiles to create
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    scale : float
        Scaling factor to use in calculations
    contact : bool
        Plot a vertical line that shows the optimal contact level. Requires
        that `pred` is a series of probabilities. Values equal to 1 (100% contact)
        will not be plotted
    prn : bool
        Print the maximum profit value (default is True, requires contact=True)
    marker : str
        Marker to use for line plot
    **kwargs : Named arguments to be passed to the seaborn lineplot function

    Returns
    -------
    Seaborn object
        Plot of profits per quantile

    Examples
    --------
    profit_plot(df, "buyer", "yes", "pred_a", cost=0.5, margin=6)
    profit_plot(df, "buyer", "yes", ["pred_a", "pred_b"], cost=0.5, margin=6)
    dct = {"Training": df.query("training == 1"), "Test": df.query("training == 0")}
    profit_plot(dct, "buyer", "yes", "pred_a", cost=0.5, margin=6)
    """
    dct = ifelse(isinstance(df, dict), df, {"": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)
    group = ifelse(len(pred) > 1 or len(dct.keys()) > 1, "predictor", None)
    df = [
        profit_tab(
            dct[k], rvar, lev, p, qnt=qnt, cost=cost, margin=margin, scale=scale
        ).assign(predictor=p + ifelse(k == "", k, f" ({k})"))
        for k in dct.keys()
        for p in pred
    ]
    df = pd.concat(df).reset_index(drop=True)
    fig = sns.lineplot(
        x="cum_prop", y="cum_profit", data=df, hue=group, marker=marker, **kwargs
    )
    fig.set(ylabel="Profit", xlabel="Percentage of population targeted")
    fig.yaxis.set_major_formatter(mtick.FuncFormatter(lambda y, _: format(int(y), ",")))
    fig.xaxis.set_major_formatter(mtick.FuncFormatter(lambda x, _: "{:.0%}".format(x)))
    fig.axhline(1, linestyle="--", linewidth=1)
    if contact:
        cnf = [
            confusion(dct[k], rvar, lev, p, cost=cost, margin=margin)[-1]
            for k in dct.keys()
            for p in pred
        ]
        prof = [
            profit_max(dct[k], rvar, lev, p, cost=cost, margin=margin, scale=scale)
            for k in dct.keys()
            for p in pred
        ]
        if prn:
            print(prof)
        [
            [
                fig.axvline(
                    l, linestyle="--", linewidth=1, color=sns.color_palette()[i]
                ),
                fig.axhline(
                    prof[i], linestyle="--", linewidth=1, color=sns.color_palette()[i]
                ),
            ]
            for i, l in enumerate(filter(lambda x: x < 1, cnf))
        ]
    if len(dct) > 1 or len(pred) > 1:
        fig.legend(title=None)

    return fig


def expected_profit_plot(
    df, rvar, lev, pred, cost=1, margin=2, scale=1, contact=True, prn=True, **kwargs
):
    """
    Plot an expected profit curve

    Parameters
    ----------
    df : Pandas dataframe or a dictionary of dataframes with keys to show multiple curves for different models or data samples
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    contact : bool
        Plot a vertical line that shows the optimal contact level. Requires
        that `pred` is a series of probabilities. Values equal to 1 (100% contact)
        will not be plotted
    prn : bool
        Print the maximum expect profit value (default is True, requires contact=True)
    **kwargs : Named arguments to be passed to the seaborn lineplot function

    Returns
    -------
    Seaborn object
        Plot of profits per quantile

    Examples
    --------
    expected_profit_plot(df, "buyer", "yes", "pred_a", cost=0.5, margin=6)
    expected_profit_plot(df, "buyer", "yes", ["pred_a", "pred_b"], cost=0.5, margin=6)
    dct = {"Training": df.query("training == 1"), "Test": df.query("training == 0")}
    expected_profit_plot(dct, "buyer", "yes", "pred_a", cost=0.5, margin=6)
    """
    dct = ifelse(isinstance(df, dict), df, {"": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)
    group = ifelse(len(pred) > 1 or len(dct.keys()) > 1, "predictor", None)

    def calc_exp_profit(df, pred, cost, margin):
        prediction = df[pred].sort_values(ascending=False)
        profit = prediction * margin - cost
        return pd.DataFrame(
            {
                "cum_prop": np.arange(1, df.shape[0] + 1) / df.shape[0],
                "cum_profit": np.cumsum(profit) * scale,
            }
        )

    df = [
        calc_exp_profit(dct[k], p, cost, margin).assign(
            predictor=p + ifelse(k == "", k, f" ({k})")
        )
        for k in dct.keys()
        for p in pred
    ]
    df = pd.concat(df).reset_index(drop=True)
    fig = sns.lineplot(x="cum_prop", y="cum_profit", data=df, hue=group, **kwargs)
    fig.set(ylabel="Expected Profit", xlabel="Percentage of population targeted")
    fig.yaxis.set_major_formatter(mtick.FuncFormatter(lambda y, _: format(int(y), ",")))
    fig.xaxis.set_major_formatter(mtick.FuncFormatter(lambda x, _: "{:.0%}".format(x)))
    fig.axhline(1, linestyle="--", linewidth=1)
    if contact:
        cnf = [
            confusion(dct[k], rvar, lev, p, cost=cost, margin=margin)[-1]
            for k in dct.keys()
            for p in pred
        ]
        eprof = df.groupby("predictor", observed=False).cum_profit.max().values
        if prn:
            print(eprof)
        [
            [
                fig.axvline(
                    l,
                    linestyle="--",
                    linewidth=1,  # color=sns.color_palette()[i]
                ),
                fig.axhline(
                    eprof[i],
                    linestyle="--",
                    linewidth=1,  # color=sns.color_palette()[i]
                ),
            ]
            for i, l in enumerate(filter(lambda x: x < 1, cnf))
        ]
    if len(dct) > 1 or len(pred) > 1:
        fig.legend(title=None)

    return fig


def ROME_plot(df, rvar, lev, pred, qnt=10, cost=1, margin=2, marker="o", **kwargs):
    """
    Plot a ROME curve

    Parameters
    ----------
    df : Pandas dataframe or a dictionary of dataframes with keys to show multiple curves for different models or data samples
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    qnt : int
        Number of quantiles to create
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    marker : str
        Marker to use for line plot
    **kwargs : Named arguments to be passed to the seaborn lineplot function

    Returns
    -------
    Seaborn object
        Plot of ROME per quantile

    Examples
    --------
    ROME_plot(df, "buyer", "yes", "pred_a", cost=0.5, margin=6)
    ROME_plot(df, "buyer", "yes", ["pred_a", "pred_b"], cost=0.5, margin=6)
    dct = {"Training": df.query("training == 1"), "Test": df.query("training == 0")}
    ROME_plot(dct, "buyer", "yes", "pred_a", cost=0.5, margin=6)
    """
    dct = ifelse(isinstance(df, dict), df, {"": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)
    group = ifelse(len(pred) > 1 or len(dct.keys()) > 1, "predictor", None)
    rd = [
        ROME_tab(dct[k], rvar, lev, p, qnt=qnt, cost=cost, margin=margin).assign(
            predictor=p + ifelse(k == "", k, f" ({k})")
        )
        for k in dct.keys()
        for p in pred
    ]
    rd = pd.concat(rd).reset_index(drop=True)
    fig = sns.lineplot(
        x="cum_prop", y="ROME", data=rd, hue=group, marker=marker, **kwargs
    )
    fig.set(
        ylabel="Return on Marketing Expenditures (ROME)",
        xlabel="Percentage of population targeted",
    )
    fig.axhline(0, linestyle="--", linewidth=1)
    if len(dct) > 1 or len(pred) > 1:
        fig.legend(title=None)
    return fig


def gains_plot(df, rvar, lev, pred, qnt=10, marker="o", **kwargs):
    """
    Plot a cumulative gains curve

    Parameters
    ----------
    df : Pandas dataframe or a dictionary of dataframes with keys to show multiple curves for different models or data samples
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column in df with model predictions
    qnt : int
        Number of quantiles to create
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    marker : str
        Marker to use for line plot
    **kwargs : Named arguments to be passed to the seaborn lineplot function

    Returns
    -------
    Seaborn object
        Plot of gaines per quantile

    Examples
    --------
    gains_plot(df, "buyer", "yes", "pred_a")
    gains_plot(df, "buyer", "yes", ["pred_a", "pred_b"], qnt=20)
    dct = {"Training": df.query("training == 1"), "Test": df.query("training == 0")}
    gains_plot(dct, "buyer", "yes", "pred_a")
    """
    dct = ifelse(isinstance(df, dict), df, {"": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)
    group = ifelse(len(pred) > 1 or len(dct.keys()) > 1, "predictor", None)
    rd = [
        gains_tab(dct[k], rvar, lev, p, qnt=qnt).assign(
            predictor=p + ifelse(k == "", k, f" ({k})")
        )
        for k in dct.keys()
        for p in pred
    ]
    rd = pd.concat(rd).reset_index(drop=True)
    fig = sns.lineplot(
        x="cum_prop", y="cum_gains", data=rd, hue=group, marker=marker, **kwargs
    )
    fig.set(ylabel="Percentage Buyers", xlabel="Percentage of population targeted")
    plt.plot([0, 1], [0, 1], linestyle="--", linewidth=1, color=plt.cm.Blues(0.7))
    fig.yaxis.set_major_formatter(mtick.FuncFormatter(lambda y, _: "{:.0%}".format(y)))
    fig.xaxis.set_major_formatter(mtick.FuncFormatter(lambda x, _: "{:.0%}".format(x)))
    if len(dct) > 1 or len(pred) > 1:
        fig.legend(title=None)
    return fig


def lift_plot(df, rvar, lev, pred, qnt=10, marker="o", **kwargs):
    """
    Plot a cumulative lift chart

    Parameters
    ----------
    df : Pandas dataframe or a dictionary of dataframes with keys to show multiple curves for different models or data samples
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name, or list, of the column(s) in df with model predictions
    qnt : int
        Number of quantiles to create
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    marker : str
        Marker to use for line plot
    **kwargs : Named arguments to be passed to the seaborn lineplot function

    Returns
    -------
    Seaborn object
        Plot of lift per quantile

    Examples
    --------
    lift_plot(df, "buyer", "yes", "pred_a")
    lift_plot(df, "buyer", "yes", ["pred_a", "pred_b"], qnt=20)
    lift = {"Training": df.query("training == 1"), "Test": df.query("training == 0")}
    lift_plot(dct, "buyer", "yes", "pred_a")
    """
    dct = ifelse(isinstance(df, dict), df, {"": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)
    group = ifelse(len(pred) > 1 or len(dct.keys()) > 1, "predictor", None)
    rd = [
        lift_tab(dct[k], rvar, lev, p, qnt=qnt).assign(
            predictor=p + ifelse(k == "", k, f" ({k})")
        )
        for k in dct.keys()
        for p in pred
    ]
    rd = pd.concat(rd).reset_index(drop=True)
    fig = sns.lineplot(
        x="cum_prop", y="cum_lift", data=rd, hue=group, marker=marker, **kwargs
    )
    fig.set(ylabel="Cumulative lift", xlabel="Percentage of population targeted")
    fig.axhline(1, linestyle="--", linewidth=1)
    if len(dct) > 1 or len(pred) > 1:
        fig.legend(title=None)
    return fig


def evalbin(df, rvar, lev, pred, cost=1, margin=2, scale=1, dec=3):
    """
    Evaluate binary classification models. Calculates TP, FP, TN, FN, contact, total,
    TPR, TNR, precision, Fscore, accuracy, profit, ROME, AUC, kappa, and profit index

    Parameters
    ----------
    df : Pandas dataframe or a dictionary of dataframes with keys to show results for
        multiple model predictions and datasets (training and test)
    rvar : str
        Name of the response variable column in df
    lev : str
        Name of the 'success' level in rvar
    pred : str
        Name of the column, of list of column names, in df with model predictions
    cost : int
        Cost of an action
    margin : int
        Benefit of an action if a successful outcome results from the action
    dec : int
        Number of decimal places to use in rounding

    Examples
    --------
    """

    dct = ifelse(isinstance(df, dict), df, {"All": df})
    pred = ifelse(isinstance(pred, str), [pred], pred)

    def calculate_metrics(key, dfm, pm):
        TP, FP, TN, FN, contact = confusion(dfm, rvar, lev, pm, cost, margin)
        total = TN + FN + FP + TP
        TPR = TP / (TP + FN)
        TNR = TN / (TN + FP)
        precision = TP / (TP + FP)
        profit = margin * TP - cost * (TP + FP)

        fpr, tpr, thresholds = metrics.roc_curve(dfm[rvar], dfm[pm], pos_label=lev)
        break_even = cost / margin
        gtbe = dfm[pm] > break_even
        pos = dfm[rvar] == lev
        if (cost * (TP + FP)) > 0:
            ROME = profit / (cost * (TP + FP))
        else:
            ROME = np.nan

        return pd.DataFrame().assign(
            Type=[key],
            predictor=[pm],
            TP=[TP],
            FP=[FP],
            TN=[TN],
            FN=[FN],
            total=[total],
            TPR=[TPR],
            TNR=[TNR],
            precision=[precision],
            Fscore=[2 * (precision * TPR) / (precision + TPR)],
            accuracy=[(TP + TN) / total],
            kappa=[metrics.cohen_kappa_score(pos, gtbe)],
            profit=[profit * scale],
            index=[0],
            ROME=[ROME],
            contact=[contact],
            AUC=[metrics.auc(fpr, tpr)],
        )

    result = pd.DataFrame()
    for key, val in dct.items():
        for p in pred:
            result = pd.concat([result, calculate_metrics(key, val, p)], axis=0)

    result.index = range(result.shape[0])
    result["index"] = result.groupby("Type", observed=False).profit.transform(
        lambda x: x / x.max()
    )
    return result.round(dec)


def auc(rvar, pred, lev=1, weights=None):
    """
    Calculate area under the RO curve (AUC)

    Calculation adapted from https://stackoverflow.com/a/50202118/1974918

    Parameters
    ----------
    rvar : Pandas series or numpy vector
        Vector with the response variable
    pred : Pandas series or numpy vector
        Vector with model predictions
    lev : str
        Name of the 'success' level in rvar

    Returns
    -------
    float :
        AUC metric

    Examples
    --------
    auc(dvd.buy, np.random.uniform(size=20000), "yes")
    auc(dvd.buy, rsm.ifelse(dvd.buy == "yes", 1, 0), "yes")
    """

    rvar = np.array(rvar)
    if not isinstance(rvar[0], bool) or lev is not None:
        rvar = rvar == lev

    if weights is None:
        rd = np.sum(rankdata(pred)[np.logical_not(rvar)])
        n1 = np.sum(np.logical_not(rvar))
        n2 = np.sum(rvar)
    else:
        pred_df = table2data(
            pd.DataFrame({"pred": pred, "rvar": rvar, "weights": weights}), "weights"
        )
        rd = np.sum(rankdata(pred_df.pred)[np.logical_not(pred_df.rvar)])
        n1 = np.sum(weights[np.logical_not(rvar)])
        n2 = np.sum(weights[rvar])

    U = rd - n1 * (n1 + 1) / 2
    wt = U / n1 / n2
    return ifelse(wt < 0.5, 1 - wt, wt)



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette.requests import Request as StarletteRequest
from shiny import App, ui, render, reactive, Inputs, Outputs, Session, req
from pathlib import Path
import webbrowser
import nest_asyncio
import uvicorn
import signal
import os
import sys
import tempfile
import pyrsm as rsm
import pyrsm.radiant.utils as ru
import pyrsm.radiant.model.model_utils as mu

choices = {
    "None": "None",
    "pred": "Prediction plots",
    "pdp": "Partial Dependence plots (PDP)",
    "vimp": "Permutation importance",
}


def summary_extra(self):
    max_features = {"sqrt": "sqrt", "log2": "log2"}
    max_features.update({v: v for v in range(1, 101, 1)})
    return (
        (
            ui.input_radio_buttons(
                "mod_type",
                None,
                selected=self.state.get("mod_type", "classification"),
                choices=["classification", "regression"],
                inline=True,
            ),
            ru.make_side_by_side(
                ui.input_numeric(
                    "n_estimators",
                    "# trees:",
                    value=self.state.get("n_estimators", 100),
                ),
                ui.input_select(
                    "max_features",
                    "Max features:",
                    choices=max_features,
                    selected=self.state.get("max_features", None),
                ),
            ),
            ru.make_side_by_side(
                ui.input_numeric(
                    "min_samples_leaf",
                    "Min node size:",
                    value=self.state.get("min_samples_leaf", 1),
                ),
                ui.input_numeric(
                    "max_samples",
                    "Sample fraction:",
                    value=self.state.get("max_samples", 1.0),
                    min=0,
                    max=1,
                    step=0.1,
                ),
            ),
            ui.input_numeric(
                "random_state",
                "Random seed:",
                value=self.state.get("random_state", 1234),
            ),
            ru.input_return_text_area(
                "extra_args",
                label="Additional arguments:",
                placeholder="max_depth=5, n_jobs=-1",
                value=self.state.get("extra_args", ""),
            ),
        ),
    )


def plots_extra(self):
    return (
        ui.panel_conditional(
            "input.plots == 'pred'",
            ui.output_ui("ui_incl_evar"),
            ui.output_ui("ui_incl_interactions"),
        ),
    )


class model_rforest:
    def __init__(
        self, datasets: dict, descriptions=None, state=None, code=True, navbar=None
    ) -> None:
        ru.init(
            self,
            datasets,
            descriptions=descriptions,
            state=state,
            code=code,
            navbar=navbar,
        )

    def shiny_ui(self, request: StarletteRequest):
        return ui.page_navbar(
            ru.head_content(),
            ui.nav_panel(
                "Model > Random Forest",
                ui.row(
                    ui.column(
                        3,
                        ru.ui_data(self),
                        ru.ui_summary(summary_extra(self)),
                        mu.ui_predict(self, show_ci=False),
                        ru.ui_plot(self, choices, plots_extra(self)),
                        style="min-width: 250px; max-width: 350px",
                    ),
                    ui.column(8, ru.ui_main_model()),
                ),
            ),
            self.navbar,
            ru.ui_help(
                "https://github.com/vnijs/pyrsm/blob/main/examples/model/model-rforest-classification.ipynb",
                "Random Forest (classification) example notebook",
            ),
            ru.ui_stop(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        # --- section standard for all apps ---
        get_data = ru.make_data_elements(self, input, output, session)

        def update_state():
            with reactive.isolate():
                ru.dct_update(self, input)

        session.on_ended(update_state)

        # --- section standard for all model apps ---
        ru.reestimate(input)

        # --- section unique to each app ---
        mu.make_model_inputs(
            self,
            input,
            output,
            get_data,
            {"classification": "isBin", "regression": "isNum"},
        )

        @output(id="ui_lev")
        @render.ui
        def ui_lev():
            req(input.rvar())
            levs = list(get_data()["data"][input.rvar()].unique())
            return ui.input_select(
                id="lev",
                label="Choose level:",
                selected=self.state.get("lev", None),
                choices=levs,
            )

        # needed for prediction plots
        mu.make_int_inputs(self, input, output, get_data)
        show_code, estimate = mu.make_estimate(
            self, input, output, get_data, fun="rforest", ret="rf", debug=False
        )
        mu.make_summary(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="rf",
            sum_fun=rsm.model.rforest.summary,
        )
        mu.make_predict(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="rf",
            pred_fun=rsm.model.rforest.predict,
            show_ci=False,
        )
        mu.make_plot(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="rf",
        )

        # --- section standard for all apps ---
        # stops returning code if moved to utils
        @reactive.Effect
        @reactive.event(input.stop, ignore_none=True)
        async def stop_app():
            rsm.md(f"```python\n{self.stop_code}\n```")
            await session.app.stop()
            os.kill(os.getpid(), signal.SIGINT)


def rforest(
    data_dct: dict = None,
    descriptions_dct: dict = None,
    state: dict = None,
    code: bool = True,
    host: str = "",
    port: int = 8000,
    log_level: str = "warning",
    debug: bool = False,
):
    """
    Launch a Radiant-for-Python app for Random Forest
    """
    host = ru.set_host(host)
    if data_dct is None:
        data_dct, descriptions_dct = ru.get_dfs(pkg="model", name="titanic")
    rc = model_rforest(data_dct, descriptions_dct, state=state, code=code)
    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    ru.message()

    # redirect stdout and stderr to the temporary file
    if not debug:
        org_stdout = sys.stdout
        org_stderr = sys.stderr
        temp = tempfile.NamedTemporaryFile()
        temp_file = open(temp.name, "w")
        sys.stdout = temp_file
        sys.stderr = temp_file

    app = App(rc.shiny_ui, rc.shiny_server)
    www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
    app_static = StaticFiles(directory=www_dir, html=False)

    routes = [
        Mount("/www", app=app_static),
        Mount("/", app=app),
    ]

    uvicorn.run(
        Starlette(debug=debug, routes=routes),
        host=host,
        port=port,
        log_level=log_level,
    )

    if not debug:
        sys.stdout = org_stdout
        sys.stderr = org_stderr
        temp_file.close()


if __name__ == "__main__":
    # titanic, titanic_description = rsm.load_data(pkg="data", name="titanic")
    # rforest({"titanic": titanic}, {"titanic": titanic_description}, code=True)
    rforest(debug=False)



from datetime import datetime
from shiny import App, ui, reactive, render
from starlette.requests import Request as StarletteRequest

state = {"__pending_changes__": False}
print("Do we get back to global after refresh?", state)  # we don't


def app_ui(request: StarletteRequest):
    if state["__pending_changes__"]:
        # from https://gist.github.com/jcheng5/88cb05e39c44704ae89e6559130e7f80
        # by Joe Cheng (Posit)
        print("There were pending changes")
        # redirect to the current URL, including query string and hash
        print(
            "Pending if-statement called at: " + datetime.now().strftime("%H:%M:%S.%f")
        )
        return ui.tags.html(
            ui.tags.meta({"http-equiv": "refresh", "content": "0"}),
            ui.tags.script("""window.Shiny.createSocket = function() {};"""),
        )

    print("Do we get back to app_ui after browser refresh?", state)  # we do
    print("Full ui re-render at: " + datetime.now().strftime("%H:%M:%S.%f"))
    return ui.page_fluid(
        ui.navset_tab(
            ui.nav_panel(
                "Main",
                ui.input_slider("n", "N", 0, 100, state.get("n", 20)),
                ui.input_text_area("f", "Filter", value=state.get("f", "")),
                ui.input_select(
                    "var1",
                    label="Variables 1:",
                    selected=state.get("var1", None),
                    choices=["a", "b", "c"],
                    multiple=True,
                ),
                ui.output_ui("ui_var2"),
                ui.div(
                    "This page was rendered at ", datetime.now().strftime("%H:%M:%S.%f")
                ),
            ),
        ),
    )


def server(input, output, session):
    def check_input_value(k, i):
        """
        The '... in input' approach causes issues (see app_state.py).
        see https://discord.com/channels/1109483223987277844/1127817202804985917/1129530385429176371
        Also tuples need to be converted to lists to be picked up on refresh by select inputs. Not
        clear why that is needed but it is
        """
        try:
            # print(f"Working on {k}")
            value = i[k]()
            if isinstance(value, tuple):
                return list(value)
            else:
                return value
        except Exception as err:
            # print(f"Extracting value for {k} failed")
            print(err)  # no error message printed
            return None

    def dct_update(input):
        input_keys = [k for k in input.__dict__["_map"].keys() if k[0] != "."]
        state.update({k: check_input_value(k, input) for k in input_keys})

    # still needs `on_flushed` so we can tell if anything changed
    def update_state():
        print("On-ended function called at: " + datetime.now().strftime("%H:%M:%S.%f"))
        with reactive.isolate():
            dct_update(input)
            state["__pending_changes__"] = False

        print(
            "On-ended function finished at: " + datetime.now().strftime("%H:%M:%S.%f")
        )

    session.on_ended(update_state)

    # session.on_flushed sets __pending_changes__ to True if any changes
    # were made. note that this does NOT update the state dictionary
    # the __pending_changes__ mechanism is needed because otherwise the
    # state update would not be completed before app_ui is re-rendered
    def on_flushed():
        print("on_flushed called")
        state["__pending_changes__"] = True

    session.on_flushed(on_flushed, once=False)

    @output(id="ui_var2")
    @render.ui
    def ui_var2():
        return ui.input_select(
            "var2",
            label="Variables 2 (dynamic):",
            selected=state.get("var2", None),
            choices=["a", "b", "c"],
            multiple=True,
        )


app = App(app_ui, server)



from pyrsm.example_data import load_data
import pandas as pd


def test_all_data():
    data, descriptions = load_data()
    assert isinstance(data, dict), "Return value should be a dictionary"
    assert len(data) == 28, "Where datasets added or removed?"
    assert [
        isinstance(v, str) for v in descriptions.values()
    ], "Descriptions should all be strings"
    assert isinstance(descriptions, dict), "Return value should be a dictionary"


def test_package():
    data, description = load_data(pkg="data")
    assert isinstance(data, dict), "Return value should be a dictionary"
    assert len(data) == 5, "Where datasets added or removed to the data package?"


def test_dataset():
    data, description = load_data(pkg="model", name="dvd")
    assert isinstance(data, pd.DataFrame), "Return value should be a dictionary"
    assert data.shape == (
        20_000,
        5,
    ), "DVD data doesn't have the right shape?"


if __name__ == "__main__":
    test_all_data()
    test_package()
    test_dataset()



import math
import matplotlib.pyplot as plt
import pandas as pd
import polars as pl
import numpy as np
import seaborn as sns
import statsmodels as sm
from sklearn.inspection import permutation_importance
from pyrsm.stats import scale_df
from pyrsm.utils import ifelse, intersect, setdiff, check_dataframe, check_series
from .model import sim_prediction, extract_evars, extract_rvar
from .perf import auc


def distr_plot(
    data: pd.DataFrame | pl.DataFrame,
    cols: list = None,
    nint: int = 25,
    bins=25,
    **kwargs,
):
    """
    Plot histograms for numeric variables and frequency plots for categorical.
    variables. Columns of type integer with less than 25 unique values will be
    treated as categorical. To change this behavior, increase or decrease the
    value of the 'nint' argument

    Parameters
    ----------
    data : Pandas dataframe
    cols: A list of column names to generate distribution plots for. If None, all
        variables will be plotted
    nint: int
        The number of unique values in a series of type integer below which the
        series will be treated as a categorical variable
    bins: int
        The number of bins to use for histograms of numeric variables
    **kwargs : Named arguments to be passed to the pandas plotting methods
    """
    data = check_dataframe(data)
    if cols is None:
        all_cols = data.columns
        cols = []
        for i, c in enumerate(all_cols):
            if (
                not pd.api.types.is_numeric_dtype(data[c].dtype)
                and not pd.api.types.is_categorical_dtype(data[c].dtype)
            ) or pd.api.types.is_object_dtype(data[c].dtype):
                print(f"No plot will be created for {c} (type {data[c].dtype})")
            else:
                cols.append(c)

    data = data.loc[:, cols].copy()

    fig, axes = plt.subplots(max(math.ceil(data.shape[1] / 2), 2), 2, figsize=(10, 2 * max(data.shape[1], 4)))
    plt.subplots_adjust(wspace=0.25, hspace=0.3)
    row = 0
    for i, c in enumerate(data.columns):
        s = data[c]
        j = ifelse(i % 2 == 0, 0, 1)
        if pd.api.types.is_integer_dtype(s.dtype) and s.nunique() < nint:
            s.sort_values().value_counts(sort=False).plot.bar(
                ax=axes[row, j], title=c, rot=0, color="slateblue", **kwargs
            )
        elif pd.api.types.is_numeric_dtype(s.dtype):
            s.plot.hist(ax=axes[row, j], title=c, rot=0, color="slateblue", bins=bins, **kwargs)
        elif pd.api.types.is_categorical_dtype(s.dtype):
            s.value_counts(sort=False).plot.bar(ax=axes[row, j], title=c, rot=0, color="slateblue", **kwargs)
        else:
            print(f"No plot will be created for {c} (type {s.dtype})")

        if j == 1:
            row += 1

    # needed because the axes object must always be 2-dimensional
    # or else the indexing used in this function will fail for small
    # DataFrames
    if data.shape[1] < 3:
        axes[-1, -1].remove()
        axes[-1, 0].remove()
        if data.shape[1] == 1:
            axes[0, -1].remove()
    elif data.shape[1] % 2 == 1:
        axes[-1, -1].remove()

    # plt.show()


def pred_plot_sm(
    fitted,
    data: pd.DataFrame | pl.DataFrame,
    incl=None,
    excl=[],
    incl_int=[],
    fix=True,
    hline=False,
    nnv=20,
    minq=0.025,
    maxq=0.975,
):
    """
    Generate prediction plots for statsmodels regression models (OLS and Logistic).
    A faster alternative to PDP plots.

    Parameters
    ----------
    fitted : A fitted (logistic) regression model
    data: Pandas or Polars DataFrame; dataset
    incl: List of strings; contains the names of the columns of data to use for prediction
          By default it will extract the names of all explanatory variables used in estimation
          Use [] to ensure no single-variable plots are created
    excl: List of strings; contains names of columns to exclude. useful when the list of
          variable names is automatically
    incl_int: List of strings; contains the names of the columns of data to be interacted for
          prediction plotting (e.g., ["x1:x2", "x2:x3"] would generate interaction plots for
          x1 and x2 and x2 and x3
    fix : Logical or tuple
        Set the desired limited on yhat or have it calculated automatically.
        Set to FALSE to have y-axis limits set by ggplot2 for each plot
    hline : Logical or float
        Add a horizontal line at the average of the target variable. When set to False
        no line is added. When set to a specific number, the horizontal line will be
        added at that value
    nnv: Integer: The number of values to simulate for numeric variables used in prediction
    minq : float
        Quantile to use for the minimum value for simulation of numeric variables
    maxq : float
        Quantile to use for the maximum value for simulation of numeric variables

    Examples
    -------
    pred_plot(lr, data, excl="monetary")
    pred_plot(lr, data, incl = [], incl_int = ["frequency:monetary"])
    """
    if hasattr(fitted, "model"):
        model = fitted.model
    else:
        # legacy for when only an un-fitted model was accepted
        model = fitted

    data = check_dataframe(data)

    # margin to add above and below in plots
    plot_margin = 0.025

    def calc_ylim(lab, lst, min_max):
        if isinstance(fix, bool) and fix:
            vals = lst[lab]
            min_vals = min(vals)
            max_vals = max(vals)
            mmin = min(min_max[0], min_vals - plot_margin * min_vals)
            mmax = max(min_max[1], max_vals + plot_margin * max_vals)
            return (mmin, mmax)
        elif not isinstance(fix, bool) and len(fix) == 2:
            return fix
        else:
            return False

    rvar = model.endog_names
    if isinstance(hline, bool):
        if hline:
            hline = data[rvar].mean()
            min_max = (hline - plot_margin * hline, hline + plot_margin * hline)
        else:
            hline = False
            min_max = (np.inf, -np.inf)
    else:
        min_max = (hline - plot_margin * hline, hline + plot_margin * hline)

    if incl is None:
        incl = extract_evars(model, data.columns)
    else:
        incl = ifelse(isinstance(incl, str), [incl], incl)

    excl = ifelse(isinstance(excl, str), [excl], excl)
    incl_int = ifelse(isinstance(incl_int, str), [incl_int], incl_int)

    if len(excl) > 0:
        incl = [i for i in incl if i not in excl]

    nr_plots = len(incl) + len(incl_int)
    if nr_plots == 0:
        return None
    fig, ax = plt.subplots(max(math.ceil(nr_plots / 2), 2), 2, figsize=(10, 2 * max(nr_plots, 4)))
    plt.subplots_adjust(wspace=0.25, hspace=0.25)

    pred_dict = {}
    for v in incl:
        iplot = sim_prediction(data, vary=v, nnv=nnv, minq=minq, maxq=maxq)
        iplot["prediction"] = fitted.predict(iplot)
        min_max = calc_ylim("prediction", iplot, min_max)
        pred_dict[v] = iplot

    for v in incl_int:
        vl = v.split(":")
        is_num = [pd.api.types.is_numeric_dtype(data[c].dtype) and data[c].nunique() > 5 for c in vl]
        iplot = sim_prediction(data, vary=vl, nnv=nnv, minq=minq, maxq=maxq)
        iplot["prediction"] = fitted.predict(iplot)
        if sum(is_num) < 2:
            min_max = calc_ylim("prediction", iplot, min_max)
        pred_dict[v] = iplot

    row = col = 0
    for i, v in enumerate(incl):
        col = ifelse(i % 2 == 0, 0, 1)
        if pd.api.types.is_numeric_dtype(data[v].dtype) and data[v].nunique() > 5:
            fig = sns.lineplot(x=v, y="prediction", data=pred_dict[v], ax=ax[row, col])
        else:
            fig = sns.lineplot(x=v, y="prediction", marker="o", data=pred_dict[v], ax=ax[row, col])
        if isinstance(min_max, tuple) and len(min_max) == 2:
            ax[row, col].set(ylim=min_max)
        if hline:
            ax[row, col].axhline(y=hline, linestyle="--")

        if col == 1:
            row += 1

    start_col = ifelse((col == 1) or len(incl) == 0, 0, 1)
    for j, v in enumerate(incl_int):
        col = ifelse(j % 2 == start_col, 0, 1)
        vl = v.split(":")
        is_num = [pd.api.types.is_numeric_dtype(data[c].dtype) and data[c].nunique() > 5 for c in vl]
        if sum(is_num) == 2:
            plot_data = (
                pred_dict[v]
                .pivot(index=vl[0], columns=vl[1], values="prediction")
                .transpose()
                .sort_values(vl[1], ascending=False)
            )
            fig = sns.heatmap(plot_data, ax=ax[row, col], xticklabels=False, yticklabels=False)
            # ax[i+j].imshow(plot_data) # not bad - investigate more
        elif sum(is_num) == 1:
            if is_num[1]:
                vl.reverse()
            fig = sns.lineplot(x=vl[0], y="prediction", hue=vl[1], data=pred_dict[v], ax=ax[row, col])
        else:
            fig = sns.lineplot(
                x=vl[0],
                y="prediction",
                hue=vl[1],
                marker="o",
                data=pred_dict[v],
                ax=ax[row, col],
            )

        if isinstance(min_max, tuple) and len(min_max) == 2 and sum(is_num) < 2:
            ax[row, col].set(ylim=min_max)
        if sum(is_num) < 2 and hline:
            ax[row, col].axhline(y=hline, linestyle="--")

        if col == 1:
            row += 1

    # needed because the axes object must always be 2-dimensional
    # or else the indexing used in this function will fail for small
    # DataFrames
    if nr_plots < 3:
        ax[-1, -1].remove()
        ax[-1, 0].remove()
        if nr_plots == 1:
            ax[0, -1].remove()
    elif nr_plots % 2 == 1:
        ax[-1, -1].remove()


def pred_plot_sk(
    fitted,
    data: pd.DataFrame | pl.DataFrame,
    rvar=None,
    incl=None,
    excl=[],
    incl_int=[],
    transformed=None,
    fix=True,
    hline=False,
    nnv=20,
    minq=0.025,
    maxq=0.975,
):
    """
    Generate prediction plots for sklearn models. A faster alternative to PDP plots
    that can handle interaction plots with categorical variables

    Parameters
    ----------
    fitted : A fitted sklearn model
    data : Polars or Pandas DataFrame with data used for estimation. Should include categorical variables
        in their original form (i.e., before using get_dummies)
    rvar : The column name for the response/target variable
    incl : A list of column names to generate prediction plots for. If None, all
        variables will be plotted
    excl : A list of column names to exclude from plotting
    incl_int : A list is ":" separated column names to plots interaction plots for.
        For example incl_int = ["a:b", "b:c"] would generate interaction plots for
        variables a x b and b x c
    transformed : List of column names that were transformed using Pandas' get_dummies. If
        None, the function will try to determine which variables might have been transformed
    fix : Logical or tuple
        Set the desired limited on yhat or have it calculated automatically.
        Set to False to have y-axis limits set by ggplot2 for each plot
    hline : Logical or float
        Add a horizontal line at the average of the target variable. When set to False
        no line is added. When set to a specific number, the horizontal line will be
        added at that value
    nnv: int
        The number of values to use in simulation for numeric variables
    minq : float
        Quantile to use for the minimum value of numeric variables
    maxq : float
        Quantile to use for the maximum value of numeric variables
    """
    # features names used in the sklearn model
    if hasattr(fitted, "feature_names_in_"):
        fn = fitted.feature_names_in_
    else:
        raise Exception(
            "This function requires a fitted sklearn model with a named features. If you are using one-hot encoding, please use get_dummies on your dataset before fitting the model."
        )

    if isinstance(data, dict):
        means = data["means"]
        stds = data["stds"]
        data = data["data"]
        data = scale_df(data, means=means, stds=stds, sf=1)
    else:
        means = None
        stds = None

    data = check_dataframe(data)

    not_transformed = [c for c in data.columns for f in fn if c == f]
    if transformed is None:
        transformed = list(
            set([c for c in data.columns for f in fn if f"{c}_" in f and c != f and f"{c}_" not in data.columns])
        )

    ints = intersect(transformed, not_transformed)
    if len(ints) > 0:
        trn = '", "'.join(transformed)
        ints = ", ".join(ints)
        not_transformed = setdiff(not_transformed, transformed)
        mess = f"""It is not clear which variables were (not) transformed using get_dummies. Please specify the
        transformed variables in your call to pred_plot_sk as `transformed=["{trn}"]`.
        In particular it is unclear how {ints} should be treated. Please remove entries from the list that
        are not correct and add any that are missing"""
        raise Exception(mess)

    if hasattr(fitted, "classes_"):
        sk_type = "classification"
    else:
        sk_type = "regression"

    def pred_fun(fitted, data):
        if sk_type == "classification":
            return fitted.predict_proba(data)[:, 1]
        else:
            if sk_type == "regression" and means is not None and stds is not None and rvar is not None:
                pred = fitted.predict(data) * stds[rvar] + means[rvar]
            else:
                pred = fitted.predict(data)

            return pred

    if incl is None:
        incl = not_transformed + transformed
    else:
        incl = ifelse(isinstance(incl, str), [incl], incl)

    def dummify(data, trs):
        if len(trs) > 0:
            return pd.concat([pd.get_dummies(data[trs], columns=trs), data.drop(trs, axis=1)], axis=1)
        else:
            return data

    # margin to add above and below in plots
    plot_margin = 0.025

    if isinstance(hline, bool):
        if hline and rvar is not None:
            hline = data[rvar].mean()
            min_max = (hline - plot_margin * hline, hline + plot_margin * hline)
        else:
            hline = False
            min_max = (np.inf, -np.inf)
    else:
        min_max = (hline - plot_margin * hline, hline + plot_margin * hline)

    def calc_ylim(lab, lst, min_max):
        if isinstance(fix, bool) and fix:
            vals = lst[lab]
            min_vals = min(vals)
            max_vals = max(vals)
            mmin = min(min_max[0], min_vals - plot_margin * min_vals)
            mmax = max(min_max[1], max_vals + plot_margin * max_vals)
            return (mmin, mmax)
        elif not isinstance(fix, bool) and len(fix) == 2:
            return fix
        else:
            return False

    excl = ifelse(isinstance(excl, str), [excl], excl)
    incl_int = ifelse(isinstance(incl_int, str), [incl_int], incl_int)

    if len(excl) > 0:
        incl = [i for i in incl if i not in excl]

    nr_plots = len(incl) + len(incl_int)
    if nr_plots == 0:
        return None

    fig, ax = plt.subplots(max(math.ceil(nr_plots / 2), 2), 2, figsize=(10, 2 * max(nr_plots, 4)))
    plt.subplots_adjust(wspace=0.25, hspace=0.25)

    pred_dict = {}
    for v in incl:
        iplot = sim_prediction(
            data[transformed + not_transformed].dropna(),
            vary=v,
            nnv=nnv,
            minq=minq,
            maxq=maxq,
        )
        iplot_dum = dummify(iplot, transformed)[fn]
        iplot["prediction"] = pred_fun(fitted, iplot_dum)
        min_max = calc_ylim("prediction", iplot, min_max)
        pred_dict[v] = iplot

    for v in incl_int:
        vl = v.split(":")
        is_num = [pd.api.types.is_numeric_dtype(data[c].dtype) and data[c].nunique() > 5 for c in vl]
        iplot = sim_prediction(
            data[transformed + not_transformed].dropna(),
            vary=vl,
            nnv=nnv,
            minq=minq,
            maxq=maxq,
        )
        iplot_dum = dummify(iplot, transformed)[fn]
        iplot["prediction"] = pred_fun(fitted, iplot_dum)
        if sum(is_num) < 2:
            min_max = calc_ylim("prediction", iplot, min_max)
        pred_dict[v] = iplot

    if len(pred_dict) > 0 and means is not None and stds is not None:
        for k in pred_dict.keys():
            for c in means.keys():
                if c in pred_dict[k].columns:
                    pred_dict[k][c] = pred_dict[k][c] * stds[c] + means[c]

    row = col = 0
    for i, v in enumerate(incl):
        col = ifelse(i % 2 == 0, 0, 1)
        if pd.api.types.is_numeric_dtype(data[v].dtype) and data[v].nunique() > 5:
            fig = sns.lineplot(x=v, y="prediction", data=pred_dict[v], ax=ax[row, col])
        else:
            fig = sns.lineplot(x=v, y="prediction", marker="o", data=pred_dict[v], ax=ax[row, col])
        if isinstance(min_max, tuple) and len(min_max) == 2:
            ax[row, col].set(ylim=min_max)
        if hline:
            ax[row, col].axhline(y=hline, linestyle="--")

        if col == 1:
            row += 1

    start_col = ifelse((col == 1) or len(incl) == 0, 0, 1)
    for j, v in enumerate(incl_int):
        col = ifelse(j % 2 == start_col, 0, 1)
        vl = v.split(":")
        is_num = [pd.api.types.is_numeric_dtype(data[c].dtype) and data[c].nunique() > 5 for c in vl]
        if sum(is_num) == 2:
            plot_data = (
                pred_dict[v]
                .pivot(index=vl[0], columns=vl[1], values="prediction")
                .transpose()
                .sort_values(vl[1], ascending=False)
            )
            fig = sns.heatmap(plot_data, ax=ax[row, col], xticklabels=False, yticklabels=False)
            # ax[i+j].imshow(plot_data) # not bad - investigate more
        elif sum(is_num) == 1:
            if is_num[1]:
                vl.reverse()
            fig = sns.lineplot(x=vl[0], y="prediction", hue=vl[1], data=pred_dict[v], ax=ax[row, col])
        else:
            fig = sns.lineplot(
                x=vl[0],
                y="prediction",
                hue=vl[1],
                marker="o",
                data=pred_dict[v],
                ax=ax[row, col],
            )

        if isinstance(min_max, tuple) and len(min_max) == 2 and sum(is_num) < 2:
            ax[row, col].set(ylim=min_max)
        if sum(is_num) < 2 and hline:
            ax[row, col].axhline(y=hline, linestyle="--")

        if col == 1:
            row += 1

    # needed because the axes object must always be 2-dimensional
    # or else the indexing used in this function will fail for small
    # DataFrames
    if nr_plots < 3:
        ax[-1, -1].remove()
        ax[-1, 0].remove()
        if nr_plots == 1:
            ax[0, -1].remove()
    elif nr_plots % 2 == 1:
        ax[-1, -1].remove()


def vimp_plot_sm(fitted, data, rep=10, ax=None, ret=False):
    """
    Creates permutation importance plots for models estimated using the
    statsmodels library

    Parameters
    ----------
    fitted : A fitted statsmodels objects
    data : Polars or Pandas DataFrame with data used for estimation
    rep: int
        The number of times to resample and calculate the permutation importance
    ax : axis object
        Add the plot to a the axis object from a matplotlib subplot
    ret: bool
        Return the variable importance table as a sorted DataFrame
    """
    fw = None
    if hasattr(fitted, "model"):
        model = fitted.model
        if hasattr(model, "_has_freq_weights") and model._has_freq_weights:
            fw = model.freq_weights
    else:
        return "This function requires a fitted linear or logistic regression"

    rvar = extract_rvar(model, data.columns)
    evars = extract_evars(model, data.columns)
    data = data[[rvar] + evars].copy().dropna().reset_index(drop=True)

    if len(model.endog) != data.shape[0]:
        raise Exception(
            "The number of rows in the DataFrame should be the same as the number of rows in the data used to estimate the model"
        )

    def imp_calc_reg(base, pred):
        return base - pd.DataFrame({"y": model.endog, "yhat": pred}).corr().iloc[0, 1] ** 2

    def imp_calc_logit(base, pred):
        return base - auc(model.endog, pred, weights=fw)

    # Calculate the baseline performance
    if isinstance(fitted, sm.genmod.generalized_linear_model.GLMResultsWrapper):
        baseline_fit = auc(model.endog, fitted.predict(data[evars]), weights=fw)
        imp_calc = imp_calc_logit  # specifying the function to use
        xlab = "Importance (AUC decrease)"
    elif isinstance(fitted, sm.regression.linear_model.RegressionResultsWrapper):
        baseline_fit = pd.DataFrame({"y": model.endog, "yhat": fitted.predict(data[evars])}).corr().iloc[0, 1] ** 2
        imp_calc = imp_calc_reg  # specifying the function to use
        xlab = "Importance (R-square decrease)"
    else:
        return "This model type is not supported. For sklearn models use vimp_plot_sk"

    # Create a copy of the dataframe
    permuted = data.copy()

    # Initialize a dictionary to store the permutation importance values
    importance_values = {v: 0 for v in evars}

    # Iterate over each feature
    for i in range(rep):
        for feature in evars:
            permuted[feature] = data[feature].sample(frac=1, random_state=i).reset_index(drop=True)
            importance_values[feature] = importance_values[feature] + imp_calc(
                baseline_fit, fitted.predict(permuted[evars])
            )
            permuted[feature] = data[feature]

    importance_values = {k: [v / rep] for k, v in importance_values.items()}
    sorted_idx = pd.DataFrame(importance_values).transpose()
    sorted_idx = sorted_idx.sort_values(0, ascending=True)
    fig = sorted_idx.plot.barh(
        color="slateblue",
        legend=None,
        figsize=(6, max(5, len(sorted_idx) * 0.4)),
        ax=ax,
    )
    plt.xlabel(xlab)
    plt.title("Permutation Importance")

    if ret:
        sorted_idx.columns = ["Importance"]
        return sorted_idx[::-1]


def vimp_plot_sk(model, rep=5, ax=None, ret=False):
    """
    Creates permutation importance plots for models estimated using the
    sklearn library. Handles categorical variables more efficiently and
    by shuffling the variable in one step rather than converting to dummy
    variables first and then shuffling each one separately.

    Parameters
    ----------
    fitted : A fitted sklearn objects
    rep: int
        The number of times to resample and calculate permutation importance
    ax : axis object
        Add the plot to a the axis object from a matplotlib subplot
    ret: bool
        Return the variable importance table as a sorted DataFrame
    """
    rvar = model.rvar
    evars = model.evar
    data = model.data[[rvar] + evars].copy().reset_index(drop=True).dropna()

    def imp_calc_reg(base, pred):
        return base - pd.DataFrame({"y": data[rvar], "yhat": pred}).corr().iloc[0, 1] ** 2

    def imp_calc_clf(base, pred):
        return base - auc(data[rvar], pred)

    # Calculate the baseline performance
    if hasattr(model.fitted, "classes_"):
        xlab = "Importance (AUC decrease)"
        baseline_fit = auc(data[rvar], model.predict(data[evars])["prediction"])
        imp_calc = imp_calc_clf  # specifying the function to use
    else:
        baseline_fit = (
            pd.DataFrame({"y": data[rvar], "yhat": model.predict(data[evars])["prediction"]}).corr().iloc[0, 1] ** 2
        )
        imp_calc = imp_calc_reg  # specifying the function to use
        xlab = "Importance (R-square decrease)"

    # Create a copy of the dataframe
    permuted = data[evars].copy()

    # Initialize a dictionary to store the permutation importance values
    importance_values = {v: 0 for v in evars}

    # Iterate over each feature
    for i in range(rep):
        for feature in evars:
            permuted[feature] = data[feature].sample(frac=1, random_state=i).reset_index(drop=True)
            importance_values[feature] += imp_calc(baseline_fit, model.predict(permuted)["prediction"])
            permuted[feature] = data[feature]  # reverting to original values

    importance_values = {k: [v / rep] for k, v in importance_values.items()}
    sorted_idx = pd.DataFrame(importance_values).transpose()
    sorted_idx = sorted_idx.sort_values(0, ascending=True)
    fig = sorted_idx.plot.barh(
        color="slateblue",
        legend=None,
        figsize=(6, max(5, len(sorted_idx) * 0.4)),
        ax=ax,
    )
    plt.xlabel(xlab)
    plt.title("Permutation Importance")

    if ret:
        sorted_idx.columns = ["Importance"]
        return sorted_idx[::-1]


def vimp_plot_sklearn(fitted, X, y, rep=5, ax=None, ret=False):
    """
    Creates permutation importance plots for models estimated using the
    sklearn library

    Parameters
    ----------
    fitted : A fitted sklearn object
    X : Polars or Pandas DataFrame with data containing the explanatory variables (features) used for estimation
    y : Series with data for the response variable (target) used for estimation
    rep: int
        The number of times to resample and calculate permutation importance
    ax : axis object
        Add the plot to a the axis object from a matplotlib subplot
    ret: bool
        Return the variable importance table as a sorted DataFrame
    """

    if hasattr(fitted, "classes_"):
        scoring = "roc_auc"
        xlab = "Importance (AUC decrease)"
    else:
        scoring = "r2"
        xlab = "Importance (R-square decrease)"

    # convert to (copied) pandas objects as needed
    X = check_dataframe(X)
    y = check_series(y)

    imp = permutation_importance(fitted, X, y, scoring=scoring, n_repeats=rep, random_state=1234)
    data = pd.DataFrame(imp.importances.T)
    data.columns = fitted.feature_names_in_
    sorted_idx = pd.DataFrame(data.mean().sort_values())
    fig = sorted_idx.plot.barh(
        color="slateblue",
        legend=None,
        figsize=(6, max(5, len(sorted_idx) * 0.4)),
        ax=ax,
    )
    fig = fig.set(title="Permutation Importance", xlabel=xlab, ylabel=None)

    if ret:
        sorted_idx.columns = ["Importance"]
        return sorted_idx[::-1]



from shiny import App, ui, reactive, render
from starlette.applications import Starlette
from starlette.routing import Mount
import pandas as pd
from starlette.requests import Request as StarletteRequest


def check_input_value(k, i):
    """
    The '... in input' approach causes issues (see app_state.py).
    see https://discord.com/channels/1109483223987277844/1127817202804985917/1129530385429176371
    Also tuples need to be converted to lists to be picked up on refresh by select inputs. Not
    clear why that is needed but it is
    """
    try:
        # print(f"Working on {k}")
        value = i[k]()
        if isinstance(value, tuple):
            return list(value)
        else:
            return value
    except Exception as err:
        # print(f"Extracting value for {k} failed")
        print(err)  # no error message printed
        return None


def dct_update(self, input):
    input_keys = [k for k in input.__dict__["_map"].keys() if k[0] != "."]
    self.state.update({k: check_input_value(k, input) for k in input_keys})


def radiant_navbar():
    return (
        ui.nav_control(
            ui.input_action_link("data", "Data", onclick='window.location.href = "/";')
        ),
        ui.nav_control(
            ui.input_action_link(
                "sa",
                "Sub App",
                onclick='window.location.href = "/sub-app/";',
            ),
        ),
    )


class radiant_data:
    def __init__(self, data, state=None) -> None:
        if state is None:
            self.state = {}
        else:
            self.state = state
        self.data = data

    def shiny_ui(self, request: StarletteRequest):
        return ui.page_navbar(
            ui.nav_panel(
                None,
                ui.row(
                    ui.column(
                        3,
                        ui.input_text_area(
                            "data_filter",
                            "Data Filter:",
                            rows=2,
                            value=self.state.get("data_filter", ""),
                            placeholder="Provide a filter (e.g., a > 1) and press return",
                        ),
                    ),
                    ui.column(8, ui.output_data_frame("show_data")),
                ),
            ),
            radiant_navbar(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input, output, session):
        def update_state():
            with reactive.isolate():
                dct_update(self, input)

        session.on_ended(update_state)

        @reactive.Calc
        def get_data():
            if input.data_filter() != "":
                try:
                    filtered_data = self.data.query(input.data_filter())
                    return filtered_data
                except Exception:
                    return self.data
            else:
                return self.data

        # self.get_data will be accessible to sub-apps as an
        # attribute of the radiant_data class instance
        # but *only* after the user has visited the Data tab
        # and self.get_data has been intialized
        self.get_data = get_data

        @output(id="show_data")
        @render.data_frame
        def show_data():
            return render.DataTable(self.get_data())


class radiant_sub_app:
    def __init__(self, data=None, state=None) -> None:
        if state is None:
            self.state = {}
        else:
            self.state = state

        if data is not None:
            self.data = data

    def shiny_ui(self, request: StarletteRequest):
        return ui.page_navbar(
            ui.nav_panel(
                None,
                ui.row(
                    ui.column(
                        3,
                        ui.input_slider("n", "N", 0, 100, self.state.get("n", 20)),
                        ui.input_select(
                            "var1",
                            label="Variables 1:",
                            selected=self.state.get("var1", None),
                            choices=["a", "b", "c"],
                            size=3,
                            multiple=True,
                        ),
                        ui.output_ui("ui_var2"),
                    ),
                    ui.column(8, ui.output_data_frame("show_data")),
                ),
            ),
            radiant_navbar(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input, output, session):
        def update_state():
            with reactive.isolate():
                dct_update(self, input)

        session.on_ended(update_state)

        @output(id="ui_var2")
        @render.ui
        def ui_var2():
            return ui.input_select(
                "var2",
                label="Variables 2 (dynamic):",
                selected=self.state.get("var2", None),
                choices=["a", "b", "c"],
                size=3,
                multiple=True,
            )

        @output(id="show_data")
        @render.data_frame
        def show_data():
            if hasattr(rc, "get_data"):
                return render.DataTable(rc.get_data())
            else:
                return render.DataTable(rc.data)


data = pd.DataFrame().assign(
    a=[1, 2, 3, 4], b=[4, 5, 6, 7], c=["w", "x", "y", "z"], d=[100, -100, 0, -1]
)

rc = radiant_data(data, state=None)
rc_sub = radiant_sub_app(data=None, state=None)

routes = [
    Mount("/sub-app", app=App(rc_sub.shiny_ui, rc_sub.shiny_server)),
    Mount("/", app=App(rc.shiny_ui, rc.shiny_server)),
]
app = Starlette(debug=True, routes=routes)



from cmath import sqrt
import matplotlib.pyplot as plt
import pandas as pd
import polars as pl
from pyrsm.utils import ifelse, check_dataframe
from pyrsm.radiant import utils as ru
from typing import Optional, Union
from scipy.stats import chisquare


class goodness:
    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        var: str,
        probs: Optional[tuple[float, ...]] = None,
        figsize: tuple[float, float] = None,
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"

        self.data = check_dataframe(self.data)
        self.var = var
        self.figsize = figsize
        self.probs = probs

        self.freq = self.data[self.var].value_counts().to_dict()
        self.nlev = len(self.freq)
        if self.probs is None:
            self.probs = [1 / self.nlev] * self.nlev

        self.observed = pd.DataFrame(
            {k: [v] for k, v in self.freq.items()},
            columns=sorted(self.freq.keys()),
        )
        self.observed["Total"] = self.observed[list(self.observed.columns)].sum(axis=1)
        self.expected = pd.DataFrame(
            {
                sorted(self.freq.keys())[i]: [
                    self.probs[i] * self.observed.at[0, "Total"],
                ]
                for i in range(len(self.freq.keys()))
            },
            columns=sorted(self.freq.keys()),
        )
        self.expected["Total"] = self.expected[list(self.expected.columns)].sum(axis=1)
        self.chisq = pd.DataFrame(
            {
                column: [
                    round(
                        (
                            (self.observed.at[0, column] - self.expected.at[0, column])
                            ** 2
                        )
                        / self.expected.at[0, column],
                        2,
                    ),
                ]
                for column in self.expected.columns.tolist()[:-1]
            },
            columns=self.expected.columns.tolist(),
        )
        self.chisq["Total"] = self.chisq[list(self.chisq.columns)].sum(axis=1)

        self.stdev = pd.DataFrame(
            {
                column: [
                    round(
                        (self.observed.at[0, column] - self.expected.at[0, column])
                        / sqrt(self.expected.at[0, column]).real,
                        2,
                    ),
                ]
                for column in self.expected.columns.tolist()[:-1]
            },
            columns=self.expected.columns.tolist()[:-1],
        )

    def summary(
        self, output: list[str] = ["observed", "expected"], dec: int = 3
    ) -> None:
        pd.set_option("display.max_columns", 20)
        pd.set_option("display.max_rows", 20)

        output = ifelse(isinstance(output, str), [output], output)

        print("Goodness of fit test")
        print(f"Data         : {self.name}")
        if self.var not in self.data.columns:
            raise ValueError(f"{self.var} does not exist in chosen dataset")

        print(f"Variable     : {self.var}")
        if self.nlev != len(self.probs):
            raise ValueError(
                f'Number of elements in "probs" should match the number of levels in {self.var} ({self.nlev})'
            )

        if not 0.999 <= sum(self.probs) <= 1.001:
            raise ValueError("Probabilities do not sum to 1 ({sum(self.probs)})")

        print(f'Probabilities: {" ".join(map(str, self.probs))}')
        print(
            f"Null hyp.    : The distribution of {self.var} is consistent with the specified distribution"
        )
        print(
            f"Alt. hyp.    : The distribution of {self.var} is not consistent with the specified distribution"
        )

        if "observed" in output:
            print("\nObserved:")
            print(self.observed.to_string(index=False))

        if "expected" in output:
            print("\nExpected: total x p")
            print(self.expected.to_string(index=False))

        if "chisq" in output:
            print("\nContribution to chi-squared: (observed - expected) ^ 2 / expected")
            print(self.chisq.to_string(index=False))

        if "dev_std" in output:
            print("\nDeviation standardized: (observed - expected) / sqrt(expected)\n")
            print(self.stdev.to_string(index=False))

        chisq, p_val = chisquare(
            [self.freq[key] for key in sorted(self.freq.keys())],
            [self.expected.at[0, key] for key in sorted(self.freq.keys())],
        )

        if p_val < 0.001:
            p_val = "< .001"
        else:
            p_val = round(p_val, dec)
        print(
            f"\nChi-squared: {round(chisq, dec)} df ({self.nlev - 1}), p.value {p_val}"
        )

    def plot(self, plots: list[str] = [], **kwargs) -> None:
        plots = ifelse(isinstance(plots, str), [plots], plots)

        args = {"rot": False}
        if "observed" in plots:
            tab = self.observed.copy().drop(columns="Total").transpose()
            args["title"] = "Observed frequencies"
            args.update(**kwargs)
            fig = tab.plot.bar(**args, legend=False)
        if "expected" in plots:
            tab = self.expected.copy().drop(columns="Total").transpose()
            args["title"] = "Expected frequencies"
            args.update(**kwargs)
            fig = tab.plot.bar(**args, legend=False)
        if "chisq" in plots:
            tab = self.chisq.drop(columns="Total").transpose()
            args["title"] = "Contribution to chi-squared statistic"
            args.update(**kwargs)
            fig = tab.plot.bar(**args, legend=False)
        if "dev_std" in plots:
            tab = self.stdev.transpose()
            args["title"] = "Deviation standardized"
            args.update(**kwargs, legend=False)
            fig, ax = plt.subplots()
            tab.plot.bar(**args, ax=ax)
            ax.axhline(y=1.96, color="black", linestyle="--")
            ax.axhline(y=1.64, color="black", linestyle="--")
            ax.axhline(y=0, color="black")
            ax.axhline(y=-1.96, color="black", linestyle="--")
            ax.axhline(y=-1.64, color="black", linestyle="--")
            ax.annotate("95%", xy=(0, 2.1), va="bottom", ha="center")
            ax.annotate("90%", xy=(0, 1.4), va="top", ha="center")


if __name__ == "__main__":
    import pyrsm as rsm

    data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="newspaper")
    gf = rsm.basics.goodness(data=data_dct, var="Income", probs=[1 / 2, 1 / 2])
    gf.summary(output=["observed"])
    gf.plot(plots=["observed", "expected", "chisq", "dev_std"])



from cmath import sqrt
import pandas as pd
import polars as pl
import numpy as np
from scipy import stats
from pyrsm.model.model import sig_stars
from pyrsm.utils import ifelse, check_dataframe
import pyrsm.basics.utils as bu
from typing import Union


class single_prop:
    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        var: str,
        lev: str = None,
        alt_hyp: str = "two-sided",
        conf: float = 0.95,
        comp_value: float = 0.5,
        test_type: str = "binomial",
    ) -> None:
        if comp_value == 0 or comp_value == 1:
            raise Exception("Please choose a comparison value between 0 and 1")

        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name].copy()
        else:
            self.data = data
            self.name = "Not provided"
        self.data = check_dataframe(self.data)
        self.var = var
        self.lev = lev
        self.alt_hyp = alt_hyp
        self.conf = conf
        self.alpha = 1 - self.conf
        self.comp_value = comp_value
        self.test_type = test_type
        self.ns = len(self.data[self.data[self.var] == self.lev])
        self.n_missing = self.data[self.var].isna().sum()
        self.n = len(self.data) - self.n_missing
        self.p = self.ns / self.n
        self.sd = sqrt(self.p * (1 - self.p)).real
        self.se = (self.sd / sqrt(self.n)).real
        self.se_p0 = sqrt(self.comp_value * (1 - self.comp_value) / self.n).real
        self.z_critical = stats.norm.ppf(1 - self.alpha / 2)
        self.z_score = None
        self.me = self.z_critical * self.se
        self.diff = self.p - self.comp_value

        def wilson_ci(zc):
            # Wilson CI used by R's prop.test function
            z_square_n = zc**2 / self.n
            denominator = 1 + z_square_n
            lower_limit = (
                self.p
                + z_square_n / 2
                - zc * np.sqrt(self.se**2 + z_square_n / (4 * self.n))
            ) / denominator
            upper_limit = (
                self.p
                + z_square_n / 2
                + zc * np.sqrt(self.se**2 + z_square_n / (4 * self.n))
            ) / denominator
            return [lower_limit, upper_limit]

        if test_type == "binomial":
            result = stats.binomtest(self.ns, self.n, self.comp_value, self.alt_hyp)
            self.ci = result.proportion_ci(confidence_level=conf)
            self.p_val = result.pvalue
        else:
            self.z_score = (self.p - self.comp_value) / self.se_p0
            # self.z_score = (self.p - self.comp_value) / self.se_p
            p_val = stats.norm.cdf(self.z_score)
            if self.alt_hyp == "two-sided":
                self.p_val = p_val * 2
                # traditional CI calculation
                # me = self.z_critical * self.se
                # self.ci = [self.p - me, self.p + me]
                self.ci = wilson_ci(self.z_critical)
            elif self.alt_hyp == "less":
                self.p_val = p_val
                # traditional CI calculation
                # z_critical = stats.norm.ppf(self.conf)
                # me = z_critical * self.se
                # self.ci = [0, self.p + me]
                self.ci = [0, wilson_ci(stats.norm.ppf(self.conf))[1]]
            else:
                self.p_val = 1 - p_val
                # traditional CI calculation
                # z_critical = stats.norm.ppf(self.conf)
                # me = z_critical * self.se
                # self.ci = [self.p - me, 1]
                self.ci = [wilson_ci(stats.norm.ppf(self.conf))[0], 1]

            # can't replicate R's z.value with statsmodels
            # var_p0 = self.comp_value * (1 - self.comp_value)
            # self.z_score, self.p_val = proportion.proportions_ztest(
            #     self.ns, self.n, self.comp_value, prop_var=var_p0
            # )

    def summary(self, dec=3) -> None:
        print(
            f'Single proportion ({ifelse(self.test_type=="binomial", "binomial exact", "z-test")})'
        )
        print(f"Data      : {self.name}")
        print(f"Variable  : {self.var}")
        print(f'Level     : "{self.lev}" in {self.var}')
        print(f"Confidence: {self.conf}")
        print(
            f'Null hyp. : the proportion of "{self.lev}" in {self.var} is equal to {self.comp_value}'
        )

        if self.alt_hyp == "less":
            alt_hyp = "less than"
        elif self.alt_hyp == "two-sided":
            alt_hyp = "not equal to"
        else:
            alt_hyp = "greater than"

        cl = bu.ci_label(self.alt_hyp, self.conf, dec=dec)

        print(
            f'Alt. hyp. : the proportion of "{self.lev}" in {self.var} {alt_hyp} {self.comp_value}\n'
        )

        stats_df = pd.DataFrame(
            {
                "p": [self.p],
                "ns": [self.ns],
                "n": [self.n],
                "n_missing": [self.n_missing],
                "sd": [self.sd],
                "se": [self.se],
                "me": [self.me],
            }
        ).round(dec)

        statistic = [
            ifelse(self.test_type == "binomial", "ns", "z.value"),
            ifelse(self.test_type == "binomial", [self.ns], [self.z_score]),
        ]
        test_df = pd.DataFrame(
            {
                "diff": [self.diff],
                statistic[0]: statistic[1],
                "p.value": [ifelse(self.p_val < 0.001, "< .001", self.p_val)],
                cl[0]: [self.ci[0]],
                cl[1]: [self.ci[1]],
                "": [sig_stars([self.p_val])[0]],
            }
        ).round(dec)

        print(stats_df.to_string(index=False), "\n")
        print(test_df.to_string(index=False))
        print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")

    def plot(self, plots: str = "bar") -> None:
        if plots == "bar":
            proportion_data = self.data[self.var].value_counts(normalize=True)
            ax = proportion_data.plot(kind="bar", alpha=0.5)
            ax.set_ylabel("")
            ax.set_title(f'Single proportion: "{self.lev}" in {self.var}')
            ax.yaxis.set_major_formatter(lambda x, _: "{:.0%}".format(x))


if __name__ == "__main__":
    import pyrsm as rsm

    consider, consider_description = rsm.load_data(pkg="basics", name="consider")
    sp = single_prop(
        data={"consider": consider},
        var="consider",
        lev="yes",
        alt_hyp="less",
        conf=0.95,
        comp_value=0.1,
        test_type="binomial",
    )
    sp.summary()
    sp.plot()



from pyrsm.stats import (
    varprop,
    seprop,
    weighted_mean,
    weighted_sd,
    scale_df,
)
import numpy as np
import pandas as pd


def test_varprop():
    assert varprop([1, 1, 1, 0, 0, 0]) == 0.25, "Proportion standard error incorrect"


def test_seprop():
    assert (
        seprop([1, 1, 1, 0, 0, 0]) == 0.2041241452319315
    ), "Proportion standard error incorrect"


# create example df and wt vector for testing
df = pd.DataFrame(
    {
        "x": [0, 1, 1, 1, 0, 0, 0],
        "y": [2, 1, 1, 1, 6, 2, 10],
        "z": [2, 1, 1, 1, 2, 2, 10],
    }
)
wt = np.array([1, 10, 1, 10, 1, 10, 1])


def test_weighted_mean():
    assert all(
        weighted_mean(df, wt).round(5) == np.array([0.61765, 1.73529, 1.617650])
    ), "Weighted means incorrect"


def test_weighted_sd():
    assert all(
        weighted_sd(df, wt).round(5) == np.array([0.48596, 1.70309, 1.53421])
    ), "Weighted standard deviations incorrect"


def test_scale_df():
    assert all(
        scale_df(df, ddof=1).round(5).loc[0, ["x", "y", "z"]].values
        == np.array([0, -0.18632, -0.10984])
    ), "Scaled pandas dataframe incorrect"
    assert all(
        scale_df(df, ddof=1).round(5).loc[1, ["x", "y", "z"]].values
        == np.array([1, -0.33123, -0.26362])
    ), "Scaled pandas dataframe incorrect"


def test_weighted_scale_df():
    assert all(
        scale_df(df, wt, ddof=0).round(5).loc[0, ["x", "y", "z"]].values
        == np.array([0, 0.07771, 0.12461])
    ), "Weighted scaled pandas dataframe incorrect"
    assert all(
        scale_df(df, wt, ddof=0).round(5).loc[1, ["x", "y", "z"]].values
        == np.array([1, -0.21587, -0.20129])
    ), "Weighted scaled pandas dataframe incorrect"


# def test_correlation():
#     cr, cp = correlation(df, prn=False)
#     assert cr[1, 0].round(3) == -0.493, "Correlations incorrect"
#     df_nan = df.copy()
#     df_nan.loc[4, "x"] = np.nan
#     cr, cp = correlation(df_nan, prn=False)
#     assert cr[1, 0].round(3) == -0.567, "Correlations with np.nan incorrect"

if __name__ == "__main__":
    test_varprop()
    test_seprop()
    test_weighted_mean()
    test_weighted_sd()
    test_scale_df()
    test_weighted_scale_df()
    # test_correlation()



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import polars as pl
from scipy import stats
from pyrsm.model.model import sig_stars
from pyrsm.utils import ifelse, check_dataframe
import pyrsm.basics.utils as bu
from typing import Literal


class single_mean:
    """
    A class to perform single-mean hypothesis testing

    Attributes
    ----------
    data : pd.DataFrame | pl.DataFrame
        The input data for the hypothesis test as a Pandas or Polars DataFrame. If a dictionary is provided, the key should be the name of the dataframe.
    var : str
        The variable/column name to test.
    alt_hyp : str
        The alternative hypothesis ('two-sided', 'greater', 'less').
    conf : float
        The confidence level for the test.
    comp_value : float
        The comparison value for the test.
    t_val : float
        The t-statistic value.
    p_val : float
        The p-value of the test.
    ci : tuple
        The confidence interval of the test.
    mean : float
        The mean of the variable.
    n : int
        The number of observations.
    n_missing : int
        The number of missing observations.
    sd : float
        The standard deviation of the variable.
    se : float
        The standard error of the variable.
    me : float
        The margin of error.
    diff : float
        The difference between the mean and the comparison value.
    df : int
        The degrees of freedom.

    Methods
    -------
    __init__(data, var, alt_hyp='two-sided', conf=0.95, comp_value=0)
        Initializes the single_mean class with the provided data and parameters.
    summary(dec=3)
        Prints a summary of the hypothesis test.
    plot()
        Plots the results of the hypothesis test.
    """
    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        var: str,
        alt_hyp: str = "two-sided",
        conf: float = 0.95,
        comp_value: float = 0,
    ):
        """
        Constructs all the necessary attributes for the single_mean object.

        Parameters
        ----------
        data : pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame]
            The input data for the hypothesis test as a Pandas or Polars DataFrame.
        var : str
            The variable/column name to test.
        alt_hyp : str, optional
            The alternative hypothesis ('two-sided', 'greater', 'less') (default is 'two-sided').
        conf : float, optional
            The confidence level for the test (default is 0.95).
        comp_value : float, optional
            The comparison value for the test (default is 0).
        """
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name].copy()
        else:
            self.data = data
            self.name = "Not provided"

        self.data = check_dataframe(self.data)
        self.var = var
        self.alt_hyp = alt_hyp
        self.conf = conf
        self.comp_value = comp_value

        result = stats.ttest_1samp(
            a=self.data[self.var],
            popmean=self.comp_value,
            nan_policy="omit",
            alternative=self.alt_hyp,
        )

        self.t_val, self.p_val = result.statistic, result.pvalue
        self.ci = result.confidence_interval(confidence_level=conf)

        self.mean = np.nanmean(self.data[self.var])
        self.n = len(self.data[self.var])
        self.n_missing = self.data[self.var].isna().sum()

        self.sd = self.data[self.var].std()
        self.se = self.data[self.var].sem()
        tscore = stats.t.ppf((1 + self.conf) / 2, self.n - 1)

        self.me = (tscore * self.se).real
        self.diff = self.mean - self.comp_value
        self.df = self.n - 1

    def summary(self, dec: int = 3) -> None:
        """
        Prints a summary of the hypothesis test.

        Parameters
        ----------
        dec : int, optional
            The number of decimal places to display (default is 3).
        """
        print("Single mean test")
        print(f"Data      : {self.name}")
        print(f"Variables : {self.var}")
        print(f"Confidence: {self.conf}")
        print(f"Comparison: {self.comp_value}\n")
        print(f"Null hyp. : the mean of {self.var} is equal to {self.comp_value}")

        if self.alt_hyp == "less":
            alt_hyp = "less than"
        elif self.alt_hyp == "two-sided":
            alt_hyp = "not equal to"
        else:
            alt_hyp = "greater than"

        cl = bu.ci_label(self.alt_hyp, self.conf, dec=dec)

        print(f"Alt. hyp. : the mean of {self.var} is {alt_hyp} {self.comp_value}\n")

        row1 = [[self.mean, self.n, self.n_missing, self.sd, self.se, self.me]]
        row2 = [
            [
                self.diff,
                self.se,
                self.t_val,
                ifelse(self.p_val < 0.001, "< .001", self.p_val),
                self.df,
                self.ci[0],
                self.ci[1],
                sig_stars([self.p_val])[0],
            ]
        ]

        col_names1 = ["mean", "n", "n_missing", "sd", "se", "me"]
        col_names2 = ["diff", "se", "t.value", "p.value", "df", cl[0], cl[1], ""]

        table1 = pd.DataFrame(row1, columns=col_names1).round(dec)
        table2 = pd.DataFrame(row2, columns=col_names2).round(dec)

        print(table1.to_string(index=False))
        print(table2.to_string(index=False))
        print("\nSignif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1")

    def plot(self, plots: Literal["hist", "sim"] = "hist") -> None:
        """
        Plots the results of the hypothesis test. If the 'hist' is selected a histogram
        of the numeric variable will be shown. The solid black line in the histogram shows 
        the sample mean. The dashed black lines show the confidence interval around the 
        sample mean. The solid red line shows the comparison value (i.e., the value under 
        the null-hypothesis). If the red line does not fall within the confidence interval 
        we can reject the null-hypothesis in favor of the alternative at the specified 
        confidence level (e.g., 0.95).

        Parameters
        ----------
        plots : str
            The type of plot to generate (default is 'hist').
 
        """
        if plots == "hist":
            fig = self.data[self.var].plot.hist(title=self.var, color="slateblue")
            plt.vlines(
                x=(self.comp_value, self.ci[0], self.mean, self.ci[1]),
                ymin=fig.get_ylim()[0],
                ymax=fig.get_ylim()[1],
                colors=("r", "k", "k", "k"),
                linestyles=("solid", "dashed", "solid", "dashed"),
            )
        elif plots == "sim":
            print("Plot type not available yet")
            # self.data[self.var].plot.hist(title=self.var, color="slateblue")
            # plt.vlines(
            #     x=[self.comp_value, self.me, self.mean],
            #     colors=["r", "k", "k"],
            #     linestyles=["solid", "dashed", "dashed"],
            # )
        else:
            print("Invalid plot type")



from starlette.applications import Starlette
from starlette.routing import Mount
from starlette.staticfiles import StaticFiles
from starlette.requests import Request as StarletteRequest
from shiny import App, render, ui, reactive, Inputs, Outputs, Session
from pathlib import Path
import webbrowser
import nest_asyncio
import uvicorn
import signal
import os
import sys
import tempfile
import pyrsm as rsm
import pyrsm.radiant.utils as ru
import pyrsm.radiant.model.model_utils as mu

choices = {
    "observed": "Observed",
    "expected": "Expected",
    "chisq": "Chi-squared",
    "dev_std": "Deviation std.",
}


def ui_summary(self):
    return ui.panel_conditional(
        "input.tabs == 'Summary'",
        ui.panel_well(
            ui.output_ui("ui_var"),
            ru.input_return_text_area(
                "probs",
                label="Probabilities:",
                placeholder="Insert list [1/2, 1/2]",
                value=self.state.get("probs", None),
            ),
            ui.input_checkbox_group(
                id="output",
                label="Select output tables:",
                selected=self.state.get("output", None),
                choices=choices,
            ),
        ),
    )


plots = {"None": "None"}
plots.update(choices)


class basics_goodness:
    def __init__(
        self, datasets: dict, descriptions=None, state=None, code=True, navbar=None
    ) -> None:
        ru.init(
            self,
            datasets,
            descriptions=descriptions,
            state=state,
            code=code,
            navbar=navbar,
        )

    def shiny_ui(self, request: StarletteRequest):
        return ui.page_navbar(
            ru.head_content(),
            ui.nav_panel(
                "Basics > Goodness-of-fit",
                ui.row(
                    ui.column(
                        3,
                        ru.ui_data(self),
                        ui_summary(self),
                        ru.ui_plot(self, plots),
                        style="min-width: 250px; max-width: 350px",
                    ),
                    ui.column(8, ru.ui_main_basics()),
                ),
            ),
            self.navbar,
            ru.ui_help(
                "https://github.com/vnijs/pyrsm/blob/main/examples/basics/basics-goodness.ipynb",
                "Goodness-of-fit example notebook",
            ),
            ru.ui_stop(),
            title="Radiant for Python",
            inverse=False,
            id="navbar_id",
        )

    def shiny_server(self, input: Inputs, output: Outputs, session: Session):
        # --- section standard for all apps ---
        get_data = ru.make_data_elements(self, input, output, session)

        def update_state():
            with reactive.isolate():
                ru.dct_update(self, input)

        session.on_ended(update_state)

        # --- section unique to each app ---
        @output(id="ui_var")
        @render.ui
        def ui_var1():
            isCat = get_data()["var_types"]["isCat"]
            return ui.input_select(
                id="var",
                label="Select a categorical variable:",
                selected=self.state.get("var", None),
                choices=isCat,
            )

        def estimation_code():
            data_name, code = (get_data()[k] for k in ["data_name", "code"])

            if ru.is_empty(input.probs()):
                probs = None
            else:
                probs = input.probs()

            args = {
                "data": f"""{{"{data_name}": {data_name}}}""",
                "var": input.var(),
                "probs": probs,
            }

            args_string = ru.drop_default_args(
                args, rsm.basics.goodness, ignore=["data", "probs"]
            )
            return f"""rsm.basics.goodness({args_string})""", code

        show_code, estimate = mu.make_estimate(
            self,
            input,
            output,
            get_data,
            fun="basics.goodness",
            ret="gf",
            ec=estimation_code,
            run=False,
            debug=True,
        )

        def summary_code():
            args = [c for c in input.output()]
            return f"""gf.summary(output={args})"""

        mu.make_summary(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="gf",
            sum_fun=rsm.basics.goodness.summary,
            sc=summary_code,
        )

        def plot_code():
            return f"""gf.plot(plots="{input.plots()}")"""

        mu.make_plot(
            self,
            input,
            output,
            session,
            show_code,
            estimate,
            ret="gf",
            pc=plot_code,
        )

        # --- section standard for all apps ---
        # stops returning code if moved to utils
        @reactive.Effect
        @reactive.event(input.stop, ignore_none=True)
        async def stop_app():
            rsm.md(f"```python\n{self.stop_code}\n```")
            await session.app.stop()
            os.kill(os.getpid(), signal.SIGINT)


def goodness(
    data_dct: dict = None,
    descriptions_dct: dict = None,
    state: dict = None,
    code: bool = True,
    host: str = "",
    port: int = 8000,
    log_level: str = "warning",
    debug: bool = False,
):
    """
    Launch a Radiant-for-Python app for goodness of fit analysis
    """
    host = ru.set_host(host)
    if data_dct is None:
        data_dct, descriptions_dct = ru.get_dfs(pkg="basics", name="newspaper")
    rc = basics_goodness(data_dct, descriptions_dct, state=state, code=code)
    nest_asyncio.apply()
    webbrowser.open(f"http://{host}:{port}")
    print(f"Listening on http://{host}:{port}")
    ru.message()

    # redirect stdout and stderr to the temporary file
    if not debug:
        org_stdout = sys.stdout
        org_stderr = sys.stderr
        temp = tempfile.NamedTemporaryFile()
        temp_file = open(temp.name, "w")
        sys.stdout = temp_file
        sys.stderr = temp_file

    app = App(rc.shiny_ui, rc.shiny_server)
    www_dir = Path(__file__).parent.parent.parent / "radiant" / "www"
    app_static = StaticFiles(directory=www_dir, html=False)

    routes = [
        Mount("/www", app=app_static),
        Mount("/", app=app),
    ]

    uvicorn.run(
        Starlette(debug=debug, routes=routes),
        host=host,
        port=port,
        log_level=log_level,
    )

    if not debug:
        sys.stdout = org_stdout
        sys.stderr = org_stderr
        temp_file.close()


if __name__ == "__main__":
    goodness(debug=False)



from typing import Optional, Literal
import pandas as pd
import polars as pl
import matplotlib.pyplot as plt
import xgboost as xgb
from sklearn.inspection import PartialDependenceDisplay as pdp
from pyrsm.utils import ifelse, check_dataframe, setdiff
from pyrsm.model.model import (
    sim_prediction,
    check_binary,
    evalreg,
    convert_to_list,
    conditional_get_dummies,
    reg_dashboard,
    nobs_dropped,
)
from pyrsm.model.perf import auc
from .visualize import pred_plot_sk, vimp_plot_sk, vimp_plot_sklearn


class xgboost:
    """
    Initialize XGBoost model

    Attributes
    ----------
    data: pandas DataFrame; dataset
    lev: String; name of the level to predict in the response variable
    rvar: String; name of the column to be used as the response variable
    evar: List of strings; contains the names of the column of data to be used as the explanatory (target) variable
    n_estimators: The number of boosting rounds
    max_depth: Maximum tree depth for base learners
    min_child_weight: Minimum sum of instance weight (hessian) needed in a child
    learning_rate: Boosting learning rate
    subsample: Subsample ratio of the training instance
    colsample_bytree: Subsample ratio of columns when constructing each tree
    random_state: Random seed used when bootstrapping data samples
    mod_type: String; type of model to be used (classification or regression)
    **kwargs : Named arguments to be passed to the XGBoost functions
    """

    def __init__(
        self,
        data: pd.DataFrame | pl.DataFrame | dict[str, pd.DataFrame | pl.DataFrame],
        rvar: Optional[str] = None,
        lev: Optional[str] = None,
        evar: Optional[list[str]] = None,
        n_estimators: int = 100,
        max_depth: int = 6,
        min_child_weight: int = 1,
        learning_rate: float = 0.3,
        subsample: float = 1.0,
        colsample_bytree: float = 1.0,
        random_state: int = 1234,
        mod_type: Literal["regression", "classification"] = "classification",
        **kwargs,
    ) -> None:
        if isinstance(data, dict):
            self.name = list(data.keys())[0]
            self.data = data[self.name]
        else:
            self.data = data
            self.name = "Not provided"
        self.data = check_dataframe(self.data)
        self.rvar = rvar
        self.lev = lev
        self.evar = convert_to_list(evar)
        self.mod_type = mod_type
        self.n_estimators = n_estimators
        self.max_depth = max_depth
        self.min_child_weight = min_child_weight
        self.learning_rate = learning_rate
        self.subsample = subsample
        self.colsample_bytree = colsample_bytree
        self.random_state = random_state
        self.ml_model = {"model": "xgboost", "mod_type": mod_type}
        self.kwargs = kwargs
        self.nobs_all = self.data.shape[0]
        self.data = self.data[[self.rvar] + self.evar].dropna()
        self.nobs = self.data.shape[0]
        self.nobs_dropped = self.nobs_all - self.nobs

        if self.mod_type == "classification":
            if self.lev is not None and self.rvar is not None:
                self.data = check_binary(self.data, self.rvar, self.lev)
            objective = "binary:logistic"
            self.xgb = xgb.XGBClassifier(
                objective=objective,
                n_estimators=self.n_estimators,
                max_depth=self.max_depth,
                min_child_weight=self.min_child_weight,
                learning_rate=self.learning_rate,
                subsample=self.subsample,
                colsample_bytree=self.colsample_bytree,
                random_state=self.random_state,
                **kwargs,
            )
        else:
            objective = "reg:squarederror"
            self.xgb = xgb.XGBRegressor(
                objective=objective,
                n_estimators=self.n_estimators,
                max_depth=self.max_depth,
                min_child_weight=self.min_child_weight,
                learning_rate=self.learning_rate,
                subsample=self.subsample,
                colsample_bytree=self.colsample_bytree,
                random_state=self.random_state,
                **kwargs,
            )

        self.data_onehot = conditional_get_dummies(self.data[self.evar])
        self.n_features = [len(evar), self.data_onehot.shape[1]]
        self.fitted = self.xgb.fit(
            self.data_onehot, self.data[self.rvar], eval_set=[(self.data_onehot, self.data[self.rvar])], verbose=False
        )

    def summary(self, dec=3) -> None:
        """
        Summarize output from an XGBoost model

        Parameters
        ----------
        dec : int, default=3
            Number of decimal places to display in the summary.

        Examples
        --------
        >>> xgb.summary()
        """
        print("XGBoost")
        print(f"Data                 : {self.name}")
        print(f"Response variable    : {self.rvar}")
        if self.mod_type == "classification":
            print(f"Level                : {self.lev}")
        print(f"Explanatory variables: {', '.join(self.evar)}")
        print(f"Model type           : {self.mod_type}")
        print(f"Nr. of features      : ({self.n_features[0]}, {self.n_features[1]})")
        print(f"Nr. of observations  : {format(self.nobs, ',.0f')}{nobs_dropped(self)}")
        print(f"n_estimators         : {self.n_estimators}")
        print(f"max_depth            : {self.max_depth}")
        print(f"min_child_weight     : {self.min_child_weight}")
        print(f"learning_rate        : {self.learning_rate}")
        print(f"subsample            : {self.subsample}")
        print(f"colsample_bytree     : {self.colsample_bytree}")
        print(f"random_state         : {self.random_state}")

        # Get predictions
        if self.mod_type == "classification":
            pred = self.fitted.predict_proba(self.data_onehot)[:, -1]
            print(f"AUC                  : {round(auc(self.data[self.rvar], pred), dec)}")
        else:
            pred=self.fitted.predict(self.data_onehot)
            print("Model fit            :")
            print(
                evalreg(
                    pd.DataFrame().assign(
                        rvar=self.data[[self.rvar]],
                        prediction=pred,
                    ),
                    "rvar",
                    "prediction",
                    dec=dec,
                )
                .T[2:]
                .rename(columns={0: " "})
                .T.to_string()
            )

        if len(self.kwargs) > 0:
            kwargs_list = [f"{k}={v}" for k, v in self.kwargs.items()]
            print(f"Extra arguments      : {', '.join(kwargs_list)}")
        print("\nEstimation data      :")
        print(self.data_onehot.head().to_string(index=False))

    def predict(self, data=None, cmd=None, data_cmd=None) -> pd.DataFrame:
        """
        Predict probabilities or values for an XGBoost model
        """

        if data is None:
            data = self.data.loc[:, self.evar].copy()
        else:
            data = data.loc[:, self.evar].copy()

        if data_cmd is not None and data_cmd != "":
            for k, v in data_cmd.items():
                data[k] = v
        elif cmd is not None and cmd != "":
            cmd = {k: ifelse(isinstance(v, str), [v], v) for k, v in cmd.items()}
            data = sim_prediction(data=data, vary=cmd)

        # only dropping the first level for binary categorical variables
        data_onehot = conditional_get_dummies(data, drop_nonvarying=False)

        # adding back levels for categorical variables is they were removed
        if data_onehot.shape[1] != self.data_onehot.shape[1]:
            for k in setdiff(self.data_onehot.columns, data_onehot.columns):
                data_onehot[k] = False
            data_onehot = data_onehot[self.data_onehot.columns]

        if self.mod_type == "classification":
            return data.assign(prediction=self.fitted.predict_proba(data_onehot)[:, -1])
        else:
            return data.assign(prediction=self.fitted.predict(data_onehot))

    def plot(
        self,
        plots: Literal["pred", "pdp", "vimp", "vimp_sklearn", "dashboard"] = "pred",
        data=None,
        incl=None,
        excl=None,
        incl_int=None,
        nobs: int = 1000,
        fix=True,
        hline=False,
        nnv=20,
        minq=0.025,
        maxq=0.975,
        figsize=None,
        ax=None,
        ret=None,
    ) -> None:
        """
        Plots for an XGBoost model

        Parameters
        ----------
        plots : str or list[str], default 'pred'
            List of plot types to generate. Options include 'pred', 'pdp', 'vimp', 'vimp_sklearn', 'dashboard'.
        nobs : int, default 1000
            Number of observations to plot. Relevant for all plots that include a scatter of data points (i.e., corr, scatter, dashboard, residual).
        incl : list[str], optional
            Variables to include in the plot. Relevant for prediction plots (pred) and coefficient plots (coef).
        excl : list[str], optional
            Variables to exclude from the plot. Relevant for prediction plots (pred) and coefficient plots (coef).
        incl_int : list[str], optional
            Interaction terms to include in the plot. Relevant for prediction plots (pred).
        fix : bool, default True
            Fix the y-axis limits. Relevant for prediction plots (pred).
        hline : bool, default False
            Add a horizontal line to the plot at the mean of the response variable. Relevant for prediction plots (pred).
        nnv : int, default 20
            Number of predicted values to calculate and to plot. Relevant for prediction plots.
        minq : float, default 0.025
            Minimum quantile of the explanatory variable values to use to calculate and plot predictions.
        maxq : float, default 0.975
            Maximum quantile of the explanatory variable values to use to calculate and plot predictions.
        figsize : tuple[int, int], default None
            Figure size for the plots in inches (e.g., "(3, 6)"). Relevant for 'corr', 'scatter', 'residual', and 'coef' plots.
        ax : plt.Axes, optional
            Axes object to plot on.
        ret : bool, optional
            Whether to return the variable (permutation) importance scores for a "vimp" plot.
        """
        plots = convert_to_list(plots)  # control for the case where a single string is passed
        excl = convert_to_list(excl)
        incl = ifelse(incl is None, None, convert_to_list(incl))
        incl_int = convert_to_list(incl_int)

        if data is None:
            data = self.data
        else:
            data = check_dataframe(data)
        if self.rvar in data.columns:
            data = data[[self.rvar] + self.evar].copy()
        else:
            data = data[self.evar].copy()

        if "pred" in plots:
            pred_plot_sk(
                self.fitted,
                data=data,
                rvar=self.rvar,
                incl=incl,
                excl=excl,
                incl_int=incl_int,
                fix=fix,
                hline=hline,
                nnv=nnv,
                minq=minq,
                maxq=maxq,
            )

        if "pdp" in plots:
            if figsize is None:
                figsize = (8, len(self.data_onehot.columns) * 2)
            fig, ax = plt.subplots(figsize=figsize)
            ax.set_title("Partial Dependence Plots")
            fig = pdp.from_estimator(self.fitted, self.data_onehot, self.data_onehot.columns, ax=ax, n_cols=2)

        if "vimp" in plots or "pimp" in plots:
            return_vimp = vimp_plot_sk(
                self,
                rep=5,
                ax=ax,
                ret=ret,
            )

            if ret:
                return return_vimp

        if "vimp_sklearn" in plots or "pimp_sklearn" in plots:
            return_vimp = vimp_plot_sklearn(
                self.fitted,
                self.data_onehot,
                self.data[self.rvar],
                rep=5,
                ax=ax,
                ret=ret,
            )

            if ret:
                return return_vimp

        if "dashboard" in plots and self.mod_type == "regression":
            model = self.fitted
            model.fittedvalues = self.predict()["prediction"]
            model.resid = self.data[self.rvar] - model.fittedvalues
            model.model = pd.DataFrame({"endog": self.data[self.rvar]})
            reg_dashboard(model, nobs=nobs)



# JSON representation of the single_mean class
single_mean_tool = {
    "name": "single_mean",
    "description": "Perform single-mean hypothesis testing.",
    "parameters": {
        "type": "object",
        "properties": {
            "data": {
                "type": "string",
                "description": "The name of the dataframe available in the environment to be used for hypothesis testing.",
            },
            "var": {"type": "string", "description": "The variable/column name to test."},
            "alt_hyp": {
                "type": "string",
                "enum": ["two-sided", "greater", "less"],
                "description": "The alternative hypothesis.",
            },
            "conf": {"type": "number", "description": "The confidence level for the test.", "default": 0.95},
            "comp_value": {"type": "number", "description": "The comparison value for the test.", "default": 0},
            "dec": {"type": "number", "description": "The number of decimals to show in the summary output", "default": 3},
            "plots": {"type": "string", "enum": ["hist", "sim"], "description": "Plots the results of the hypothesis test. If the 'hist' plot is selected a histogram of the numeric variable will be shown. The solid black line in the histogram shows the sample mean. The dashed black lines show the confidence interval around the sample mean. The solid red line shows the comparison value (i.e., the value under the null-hypothesis). If the red line does not fall within the confidence interval we can reject the null-hypothesis in favor of the alternative at the specified confidence level (e.g., 0.95). The 'sim' plot is not currently available.", "default": "hist"},
        },
        "required": ["data", "var"],
    },
}

# JSON representation of the compare_means class
compare_means_tool = {
    "name": "compare_means",
    "description": "Perform comparison of means hypothesis testing.",
    "parameters": {
        "type": "object",
        "properties": {
            "data": {
                "type": "string",
                "description": "The name of the dataframe available in the environment to be used for hypothesis testing.",
            },
            "var1": {
                "type": "string",
                "description": "The first variable/column name to test. This can be numeric or categorical.",
            },
            "var2": {"type": "string", "description": "The second variable/column name to test. This must be numeric."},
            "comb": {
                "type": "array",
                "items": {"type": "string"},
                "description": "List of string tuples representing combinations of variables for comparison. Defaults to all possible combinations.",
                "default": [],
            },
            "alt_hyp": {
                "type": "string",
                "enum": ["two-sided", "greater", "less"],
                "description": "The alternative hypothesis.",
                "default": "two-sided",
            },
            "conf": {"type": "number", "description": "The confidence level for the test.", "default": 0.95},
            "sample_type": {
                "type": "string",
                "enum": ["independent", "paired"],
                "description": "Type of samples ('independent' or 'paired').",
                "default": "independent",
            },
            "adjust": {
                "type": "string",
                "enum": [None, "bonferroni"],
                "description": "Adjustment for multiple testing (e.g., None or 'bonferroni').",
                "default": None,
            },
            "test_type": {
                "type": "string",
                "enum": ["t-test", "wilcox"],
                "description": "The type of test to use ('t-test' or 'wilcox').",
                "default": "t-test",
            },
            "dec": {
                "type": "integer",
                "default": 3,
                "description": "The number of decimal places to display in the summary.",
            },
            "extra": {
                "type": "boolean",
                "default": False,
                "description": "Whether to include extra details in the summary output.",
            },
            "plots": {
                "type": "string",
                "enum": ["scatter", "box", "density", "bar"],
                "description": "The type of plot to create ('scatter', 'box', 'density', 'bar').",
                "default": "scatter",
            },
            "nobs": {
                "type": "integer",
                "description": "The number of observations to plot. Defaults to all available datapoints.",
                "default": None,
            },
        },
        "required": ["data", "var1", "var2"],
    },
}

linear_regression_tool = {
    "name": "regress",
    "description": "Perform linear regression modeling.",
    "parameters": {
        "type": "object",
        "properties": {
            "data": {
                "type": "string",
                "description": "The name of the dataframe available in the environment to be used for linear regression analysis.",
            },
            "rvar": {
                "type": "string",
                "description": "The response or dependent variable to be used in the regression model. This must be a numeric variable. Ask the user to select this variable from a list of numeric variables.",
            },
            "evar": {
                "type": "string",
                "description": "A required list of explanatory or independent variables to be used in the regression model as a list of the variable names. If therer fewer than 10 explanatory variables in the data use all of them to start. Else, ask the user to select these variables from a list of all available variables but exclude the response variable.",
            },
            "ivar": {
                "type": "string",
                "description": "An optional list of interaction terms to included as explanatory (independent) variable in the regression model. Ask the user if they want to include interactions terms. If they do, ask them select the interactions as a pair of explanatory variables. For example, if we have variables x1 and x2 in the model as explanatory variables in the model, a possible interaction term would be included as 'x1:x2'. Notice the colon between the variable names. Do NOT include interactions unless the user explicity asks for them.",
            },
        },
        "required": ["data", "rvar", "evar"],
    },
}

tools = [
    {
        "type": "function",
        "function": single_mean_tool,
    },
    {
        "type": "function",
        "function": compare_means_tool,
    },
    {
        "type": "function",
        "function": linear_regression_tool,
    },

]
